
#%%
from typing import Union, List
from pathlib import Path
from datetime import datetime
from astropy.table import Table, vstack
from astropy.time import Time
from astropy.io import fits, ascii
from tqdm import tqdm
import re
import numpy as np
from shapely.geometry import Polygon
from astropy.nddata import Cutout2D
from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.units import Quantity
from astropy.wcs import WCS
from matplotlib.gridspec import GridSpec
from sklearn.preprocessing import MinMaxScaler
from astropy.visualization import ZScaleInterval, MinMaxInterval
import matplotlib.pyplot as plt
import uuid

from ezphot.helper import Helper
from ezphot.imageobjects import ScienceImage, ReferenceImage
from ezphot.imageobjects import Background, Errormap, Mask
from ezphot.methods import Stack
from ezphot.methods import PhotometricCalibration
from ezphot.methods import AperturePhotometry
from ezphot.methods import BackgroundGenerator
from ezphot.methods import MaskGenerator
from ezphot.dataobjects import Catalog
from ezphot.utils import ImageQuerier
from ezphot.methods import Reproject
#%%
class Subtract:
    """
    Subtract a reference image from a target image.
    
    This class will prepare the subtracted region by trimming both images to the overlapping region.
    Then, run HOTPANTS subtraction to find transients.
    """
    
    def __init__(self):
        """
        Initialize the Subtract class.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self.helper = Helper()
        self.background = BackgroundGenerator()
        self.combiner = Stack()
        self.aperphot = AperturePhotometry()
        self.photcal = PhotometricCalibration()
        self.masking = MaskGenerator()
        self.projection = Reproject()
        self.imagequerier = ImageQuerier()
        
    def subtract(self,
                 target_img: ScienceImage,
                 reference_img: ReferenceImage,
                 target_ivpmask: Mask = None,
                 reference_ivpmask: Mask = None,
                 target_stamp: str = None,
                 id_ : int = 0,
                 save: bool = True,
                 verbose: bool = True,
                 visualize: bool = True,
                 convim: str = None,
                 normim: str = 'i',
                 nrx: int = 1,
                 nry: int = 1,
                 iu: float = 60000,
                 il: float = -10000,
                 tu: float = 60000,
                 tl: float = -10000,
                 **hotpants_params):
        """
        Subtract the target_img from reference-img. 
        It will prepare the subtracted region by trimming both images to the overlapping region.
        Then, run HOTPANTS subtraction.
        
        Parameters
        ----------
        target_img : ScienceImage
            The target image to subtract from.
        reference_img : ReferenceImage
            The reference image to subtract.
        target_ivpmask : Mask, optional
            The mask for the target image.
        reference_ivpmask : Mask, optional
            The mask for the reference image.
        target_stamp : str, optional
            The stamp for the target image.
        id_ : int, optional
            The id for the target image.
        save : bool, optional
            Whether to save the subtracted image.
        verbose : bool, optional
            Whether to print verbose output.
        visualize : bool, optional
            Whether to visualize the subtracted image.
        convim : str, optional
            The convolution image type. ['t', 'i']
        normim : str, optional
            The normalization image type. ['t', 'i']
        nrx : int, optional
            The number of regions for subtraction.
        nry : int, optional
            The number of regions for subtraction.
        iu : float, optional
            The upper limit for the input image.
        il : float, optional
            The lower limit for the input image.
        tu : float, optional
            The upper limit for the template image.
        tl : float, optional
            The lower limit for the template image.
        **hotpants_params : dict, optional
            Additional keyword arguments for the HOTPANTS subtraction.
            
        Returns
        -------
        (subframe_target_img, subframe_reference_img, subframe_subtract_img, fullframe_subtract_mask, subframe_subtract_mask) : Tuple[ScienceImage, ReferenceImage, ScienceImage, Mask, Mask]
            The subtracted image, the reference image, the subtracted image, the full frame mask, and the subframe mask.
        """
        # Set the target_stamp 
        if target_stamp is not None:
            self.helper.print(f"Using target stamp from {target_stamp}", verbose)
                        
        target_seeing = target_img.seeing
        reference_seeing = reference_img.seeing
        if convim is None:
            if target_seeing is None or reference_seeing is None:
                self.helper.print("Warning: One of the images does not have a valid seeing value. Using 't' for convolution.", verbose)
                convim = 't'
            elif target_seeing > reference_seeing:
                convim = 't'
            else:
                convim = 'i'
        
        # Trim images
        subframe_target_img, subframe_reference_img, subframe_target_ivpmask, subframe_reference_ivpmask, fullframe_subtract_mask, subframe_subtract_mask, subframe_target_stamp = self._prepare_subtract_region(
            target_img = target_img,
            reference_img = reference_img,
            target_ivpmask = target_ivpmask,
            reference_ivpmask = reference_ivpmask,
            target_stamp = target_stamp,
            id_ = id_,
            save = True,
            verbose = verbose,
            visualize = visualize
        )
            
        # Run HOTPANTS subtraction
        result = self.helper.run_hotpants(
            target_path = subframe_target_img.savepath.savepath,
            reference_path = subframe_reference_img.savepath.savepath,
            convolve_path = subframe_target_img.savepath.convolvepath,
            target_mask = subframe_target_ivpmask.savepath.savepath,
            reference_mask = subframe_reference_ivpmask.savepath.savepath,
            stamp = subframe_target_stamp,
            target_outpath = subframe_target_img.savepath.subtractpath,
            convim = convim,
            normim = normim,
            nrx = nrx,
            nry = nry,
            iu = iu,
            il = il,
            tu = tu,
            tl = tl,
            **hotpants_params
        )            
        
        subframe_convolve_img = type(subframe_target_img)(subframe_target_img.savepath.convolvepath, telinfo = subframe_target_img.telinfo, load = True)
        #subframe_convolve_img.remove()
        subframe_subtract_img = type(target_img)(result, telinfo = target_img.telinfo, load = True)
        subframe_subtract_mask.path = subframe_subtract_img.savepath.invalidmaskpath
        
        if save:
            subframe_subtract_img.write()
            fullframe_subtract_mask.write()
            subframe_subtract_mask.write()
        else:
            subframe_subtract_img.data
            subframe_subtract_img.remove()
            
        if visualize:
            subframe_subtract_img.show()
        return subframe_target_img, subframe_reference_img, subframe_subtract_img, fullframe_subtract_mask, subframe_subtract_mask
    
    def find_transients(self,
                        target_img: ScienceImage,
                        reference_imglist: List[ReferenceImage],
                        target_bkg: Background = None,
                        
                        # Photometry configuration
                        detection_sigma: float = 5,
                        aperture_diameter_arcsec: List[float] = [5, 7, 10],
                        aperture_diameter_seeing: List[float] = [3.5, 4.5],
                        kron_factor: float = 1.5,
                        catalog_type: str = 'GAIAXP',
                        
                        target_transient_number: int = 5,
                        reject_variable_sources: bool = False,
                        negative_detection: bool = True,
                        reverse_subtraction: bool = False,
                        
                        save: bool = True,
                        verbose: bool = True,
                        visualize: bool = False,
                        save_transient_figure: bool = True,
                        save_candidate_figure: bool = True,
                        show_transient_numbers: int = 100,
                        show_candidate_numbers: int = 100,
                        tu: float = None,
                        tl: float = None,
                        iu: float = None,
                        il: float = None,
                        nrx: int = 1,
                        nry: int = 1,
                        nsx: int = 10,
                        nsy: int = 10,
                        ko: int = 3,
                        bgo: int = 1,
                        r: int = 10,
                        **hotpants_params):
        """
        Find transients in the subtracted image.
        
        This function will prepare the subtracted image by trimming the target image to the overlapping region.
        Then, run HOTPANTS subtraction to find transients.
        
        Parameters
        ----------
        target_img : ScienceImage
            The target image to find transients in.
        reference_imglist : List[ReferenceImage]
            The list of reference images to find transients in.
        target_bkg : Background, optional
            The background for the target image.
        detection_sigma : float, optional
            The detection sigma for the target image.
        aperture_diameter_arcsec : List[float], optional
            The aperture diameter in arcseconds for the target image.
        aperture_diameter_seeing : List[float], optional
            The aperture diameter in seeing for the target image.
        kron_factor : float, optional
            The Kron factor for the target image.
        target_transient_number : int, optional
            The number of transients to find in the target image.
        reject_variable_sources : bool, optional
            Whether to reject variable sources.
        negative_detection : bool, optional
            Whether to detect negative sources.
        reverse_subtraction : bool, optional
            Whether to reverse the subtraction.
        save : bool, optional
            Whether to save the subtracted image.
        verbose : bool, optional
            Whether to print verbose output.
        visualize : bool, optional
            Whether to visualize the subtracted image.
        save_transient_figure : bool, optional
            Whether to save the transient figure.
        save_candidate_figure : bool, optional
            Whether to save the candidate figure.
        show_transient_numbers : int, optional
            The number of transients to show in the transient figure.
        show_candidate_numbers : int, optional
            The number of candidates to show in the candidate figure.
        tu : float, optional
            The upper limit for the template image.
        tl : float, optional
            The lower limit for the input image.
        iu : float, optional
            The upper limit for the input image.
        il : float, optional
            The lower limit for the input image.
        nrx : int, optional
            The number of regions for subtraction.
        nry : int, optional
            The number of regions for subtraction.
        nsx : int, optional
            The number of regions for subtraction.
        nsy : int, optional
            The number of regions for subtraction.
        ko : int, optional
            The number of regions for subtraction.
        bgo : int, optional
            The number of regions for subtraction.
        r : int, optional
            The number of regions for subtraction.
        **hotpants_params : dict, optional
            Additional keyword arguments for the HOTPANTS subtraction.
            
        Returns
        -------
        (transient_catalogs, candidate_catalogs) : Tuple[List[Table], List[Table]]
            The list of transient catalogs and the list of candidate catalogs.
        """
        # Prepare reprojected target_img
        if target_bkg is not None:
            target_img_sub = self.background.subtract_background(
                target_img = target_img,
                target_bkg = target_bkg,
                save = save,
                overwrite = False,
                visualize = visualize,
                save_fig = False)
        else:
            target_img_sub = target_img.copy()
        
        # Free memory
        target_img.clear()
        if target_bkg is not None:
            target_bkg.clear()
                
        self.helper.print("================== Target Image Preparation ==================", verbose)
        center_target = target_img.center
        reprojected_target_img, _, reprojected_target_ivpmask = self.projection.reproject(
            target_img = target_img_sub,
            swarp_params = None,
            resample_type = 'LANCZOS3',
            center_ra = center_target['ra'],
            center_dec = center_target['dec'],
            x_size = target_img_sub.naxis1,
            y_size = target_img_sub.naxis2,
            pixelscale = target_img_sub.pixelscale.mean(),
            verbose = verbose,
            overwrite = False,
            save = save,
            return_ivpmask = True
        )
        # Free memory
        target_img_sub.clear()
        
        # Calculate typical detection criteria for transients
        reprojected_target_catalog = self.aperphot.sex_photometry(
            target_img = reprojected_target_img,
            target_bkg = None,  # No background for subtraction
            target_bkgrms = None,  # No background RMS for subtraction
            target_mask = None,
            detection_sigma = detection_sigma,
            aperture_diameter_arcsec = aperture_diameter_arcsec,
            aperture_diameter_seeing = aperture_diameter_seeing,
            saturation_level = 60000,
            kron_factor = kron_factor,
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = False
        )
        
        reprojected_target_img, reprojected_target_catalog, reprojected_target_ref_catalog = self.photcal.photometric_calibration(
            target_img = reprojected_target_img,
            target_catalog = reprojected_target_catalog,
            catalog_type = catalog_type,
            max_distance_second = 5.0,
            calculate_color_terms = False,
            save = True,
            verbose = verbose,
            visualize = visualize,
            save_fig = False,
            save_starcat = False,
            save_refcat = True
            )
        
        reprojected_target_img.remove(remove_main = False, remove_connected_files = True, skip_exts = ['.refcat', '.invalidmask'])
        
        transient_criteria = dict()
        if 'SATURATE' in reprojected_target_img.header.keys():
            saturation_level_target = min(reprojected_target_img.header['SATURATE'] * 0.5, 60000)
        else:
            saturation_level_target = 60000
        reprojected_target_ref_catalog_data = reprojected_target_ref_catalog.data
        transient_criteria['classstar_lower'] = min(0.5, 0.9 * np.median(reprojected_target_ref_catalog_data['CLASS_STAR']))
        transient_criteria['elongation_upper'] = max(1.5, 1.2 * np.median(reprojected_target_ref_catalog_data['ELONGATION']))
        if iu is None:
            transient_criteria['flux_upper_target'] = saturation_level_target
        else:
            transient_criteria['flux_upper_target'] = iu
        if il is None:
            transient_criteria['flux_lower_target'] = reprojected_target_img.info.SKYVAL - 15 * reprojected_target_img.info.SKYSIG if reprojected_target_img.info.SKYVAL is not None and reprojected_target_img.info.SKYSIG is not None else -10000
        else:
            transient_criteria['flux_lower_target'] = il
        is_criteria_determined = True
        
        reprojected_target_stamp = None
        if len(reprojected_target_ref_catalog.data) > 30:
            reprojected_target_stamp = reprojected_target_ref_catalog.to_stamp(reprojected_target_img)

        self.helper.print("=============== Target Image Preparation END ===============", verbose)
        
        self.helper.print(f"=============== Subtraction with {len(reference_imglist)} images ===============", verbose)
        final_subtraction_mask = Mask(path = target_img.savepath.submaskpath, masktype = 'subtraction', load = False)
        final_subtraction_mask.data = np.zeros_like(reprojected_target_img.data, dtype=bool)
        
        all_catalogs = []
        transient_catalogs = []
        candidate_catalogs = []
        for i, reference_img in tqdm(enumerate(reference_imglist), total=len(reference_imglist), desc="Subtraction Progress"):
            reference_img_temp = reference_img.copy()
            reference_img_temp.path = reference_img.savedir / (uuid.uuid4().hex + '.fits')
            reference_img_temp.write()
            
            # Step 2: Reproject reference images to reprojected_target_img
            reprojected_reference_img, reprojected_reference_ivpmask = self._reproject_to_target(
                reference_img = reference_img_temp,
                target_img = reprojected_target_img,
                save = save,
                verbose = verbose
            )
            
            reference_img_temp.clear()
            reference_img_temp.remove(remove_main = True, remove_connected_files = True, skip_exts = [''])
            #reference_img.clear()
            #reprojected_reference_img.remove(remove_main = False, remove_connected_files = True, skip_exts = ['.invalidmask'])
            
            # Step 1: If reference_img.seeing is None, update seeing
            reference_catalog = None
            reference_ref_catalog = None
            reprojected_reference_stamp = None
            if reference_img.seeing is None or reject_variable_sources:
                # If reference catalog is not available, run SExtractor
                reference_catalog = reference_img.catalog
                if reference_catalog is None:
                    reference_catalog = self.aperphot.sex_photometry(
                        target_img = reprojected_reference_img,
                        target_bkg = None,  # No background for subtraction
                        target_bkgrms = None,  # No background RMS for subtraction
                        target_mask = reprojected_reference_ivpmask,
                        detection_sigma = detection_sigma,
                        aperture_diameter_arcsec = aperture_diameter_arcsec,
                        aperture_diameter_seeing = aperture_diameter_seeing,
                        saturation_level = 60000,
                        kron_factor = kron_factor,
                        save = save,
                        verbose = verbose,
                        visualize = visualize,
                        save_fig = False)
                
                try:
                    reprojected_reference_img, reprojected_reference_catalog, reprojected_reference_ref_catalog = self.photcal.photometric_calibration(
                        target_img = reprojected_reference_img,
                        target_catalog = reference_catalog,
                        catalog_type = catalog_type,
                        max_distance_second = 5.0,
                        calculate_color_terms = False,
                        calculate_mag_terms = False,
                        save = True,
                        verbose = verbose,
                        visualize = visualize,
                        save_fig = False,
                        save_starcat = False,
                        save_refcat = True
                        )         
                    #reprojected_reference_stamp = reprojected_reference_ref_catalog.to_stamp(reprojected_reference_img)
                except:
                    peeing = np.median(reference_catalog.data['FWHM_IMAGE'])
                    reprojected_reference_img.header['SEEING'] = peeing * np.mean(reprojected_reference_img.pixelscale)


            if np.sum(reprojected_reference_ivpmask.data == 0) < 20000:
                self.helper.print('Reference image and target image are not overlapping enough for subtraction.', verbose)
                failed_reference_images.append(i)
                continue
            
            # Step 3. Update transient criteria based on the seeing of reprojected images
            target_seeing = reprojected_target_img.seeing
            reference_seeing = reprojected_reference_img.seeing
            # If target_seeing larger, convolve to target_img, thus the subtracted image's seeing should be target_seeing 
            if target_seeing >= reference_seeing:
                subtract_seeing = target_seeing
                sigma_match = np.sqrt(target_seeing**2 - reference_seeing**2)
            else:
                subtract_seeing = reference_seeing
                sigma_match = np.sqrt(reference_seeing**2 - target_seeing**2)
            convolved_source_minarea =  np.pi * (subtract_seeing/2/np.mean(reprojected_target_img.pixelscale))**2 + 0.3 * 2 * np.pi * (sigma_match / 2.355 / np.mean(reprojected_target_img.pixelscale))**2
                
            saturation_level_reference = 60000
            if 'SATURATE' in reprojected_reference_img.header.keys():
                saturation_level_reference = min(1.5* reprojected_reference_img.header['SATURATE'],60000)
            transient_criteria['seeing_upper'] = 1.3 * subtract_seeing
            transient_criteria['seeing_lower'] = max(1.0, 0.7 * subtract_seeing)    
            # Set saturation level
            if tu is None:
                transient_criteria['flux_upper_reference'] = saturation_level_reference
            else:
                transient_criteria['flux_upper_reference'] = tu
            if tl is None:
                transient_criteria['flux_lower_reference'] = reprojected_reference_img.info.SKYVAL - 15 * reprojected_reference_img.info.SKYSIG if reprojected_reference_img.info.SKYVAL is not None and reprojected_reference_img.info.SKYSIG is not None else -10000
            else:
                transient_criteria['flux_lower_reference'] = tl
            
            #ng = f"3 6 %.2f 4 %.2f 2 %.2f" %(0.5 * sigma_match, sigma_match, 2.0 * sigma_match)
            
            # Step 4: Subtration
            subframe_target_imglist = []
            subframe_reference_imglist = []
            subframe_subtract_imglist = []
    
            subframe_target_img, subframe_reference_img, subframe_subtract_img, fullframe_subtract_mask, subframe_subtract_ivpmask = self.subtract(
                target_img = reprojected_target_img,
                reference_img = reprojected_reference_img,
                target_ivpmask = reprojected_target_ivpmask,
                reference_ivpmask = reprojected_reference_ivpmask,
                target_stamp = reprojected_target_stamp,
                id_ = i,
                save = save,
                verbose = verbose,
                visualize = visualize,
                # HOTPANTS Parameters
                convim = None,
                normim = 'i',
                nrx = nrx,
                nry = nry,
                iu = transient_criteria['flux_upper_target'],
                il = transient_criteria['flux_lower_target'],
                tu = transient_criteria['flux_upper_reference'],
                tl = transient_criteria['flux_lower_reference'],
                # Other hotpants parameters
                ko = ko,
                bgo = bgo,
                nsx = nsx,
                nsy = nsy,
                r = r,
                **hotpants_params)
            
            reprojected_reference_img.remove(remove_main = True, remove_connected_files = True)
            subframe_target_img.remove(remove_main = False, remove_connected_files = True)
            subframe_reference_img.remove(remove_main = False, remove_connected_files = True)
            subframe_subtract_img.remove(remove_main = False, remove_connected_files = True, skip_exts = ['.invalidmask'])
            final_subtraction_mask.combine_mask(fullframe_subtract_mask.data, operation='add')
            
            output_name = subframe_target_img.savepath.savepath.name
            subframe_target_img.rename(new_name = 'sci_' + output_name)
            subframe_reference_img.rename(new_name = 'ref_'+ output_name)

            # Step 5: Extract sources
            tbl_first = self.aperphot.sex_photometry(
                target_img = subframe_subtract_img,
                target_bkg = None,  # No background for subtraction
                target_bkgrms = None,  # No background RMS for subtraction
                target_mask = subframe_subtract_ivpmask,
                sex_params = dict(SEEING_FWHM = subtract_seeing, DETECT_MINAREA = convolved_source_minarea),
                detection_sigma = detection_sigma,
                aperture_diameter_arcsec = aperture_diameter_arcsec,
                aperture_diameter_seeing = aperture_diameter_seeing,  
                saturation_level = 60000,
                save = save,
                verbose = verbose,
                visualize = visualize,
                save_fig = False
            )
            self.photcal.apply_zp(
                target_img = subframe_subtract_img,
                target_catalog = tbl_first,
                save = True
            )
        
            # Step 6: Filter the table for significant sources
            all_catalog, selected_catalog = self.select_transients(
                target_catalog = tbl_first,
                snr_lower = 5.0,
                fwhm_lower = transient_criteria['seeing_lower'],
                fwhm_upper = transient_criteria['seeing_upper'],
                flag_upper = 1,
                maskflag_upper = 1,
                class_star_lower = transient_criteria['classstar_lower'],
                elongation_upper = transient_criteria['elongation_upper'],
                flux_key = 'FLUX_AUTO',
                fluxerr_key = 'FLUXERR_AUTO',
                fwhm_key = 'FWHM_WORLD',
                flag_key = 'FLAGS',
                maskflag_key = 'IMAFLAGS_ISO',
                classstar_key = 'CLASS_STAR',
                elongation_key = 'ELONGATION',
                verbose = verbose,
                save = save,
                return_only_transient = True,
            )
            
            all_tbl = all_catalog.copy()  
            candidate_tbl = selected_catalog.copy()

            subframe_subtract_img.remove(remove_main = False, remove_connected_files = True, skip_exts = ['.invalidmask', '.transient', '.candidate', '.cat'])

            # Step 7: negative image & Photometry
            if negative_detection:
                negative_subframe_subtract_img = subframe_subtract_img.copy()
                negative_subframe_subtract_img.path = subframe_subtract_img.savepath.savedir / ('inv_' + subframe_subtract_img.savepath.savepath.name)
                negative_subframe_subtract_img.data = -subframe_subtract_img.data        

                tbl_second = self.aperphot.sex_photometry(
                    target_img = negative_subframe_subtract_img,
                    target_bkg = None,  # No background for subtraction
                    target_bkgrms = None,  # No background RMS for subtraction
                    target_mask = subframe_subtract_ivpmask,
                    sex_params = dict(SEEING_FWHM = subtract_seeing, DETECT_MINAREA = convolved_source_minarea),
                    detection_sigma = detection_sigma,
                    aperture_diameter_arcsec = aperture_diameter_arcsec,
                    aperture_diameter_seeing = aperture_diameter_seeing,    
                    saturation_level = 60000,
                    kron_factor = kron_factor,
                    save = False,
                    verbose = True,
                    visualize = visualize,
                    save_fig = False
                )

                # Remove
                negative_subframe_subtract_img.remove()
                
                # Only keep the unmatched sources from the first photometry step
                if len(tbl_second.data) > 0:
                    coord_first = SkyCoord(ra=candidate_tbl.data['X_WORLD'],
                                        dec=candidate_tbl.data['Y_WORLD'],
                                        unit = 'deg')
                    coord_second = SkyCoord(ra=tbl_second.data['X_WORLD'],
                                            dec=tbl_second.data['Y_WORLD'],
                                            unit = 'deg')
                    matched_first, matched_second, unmatched_first = self.helper.cross_match(coord_first, coord_second, subtract_seeing)
                    candidate_tbl.data = candidate_tbl.data[unmatched_first]
                    self.helper.print(f"Found {len(candidate_tbl.data)} transients after negative detection.", verbose)
                else:
                    self.helper.print("No significant sources found in the second photometry step.", verbose)
                    pass
                
            # Match the first photometry results with the reference catalog to reject variable sources
            transient_tbl = None
            if reject_variable_sources:
                transient_tbl = candidate_tbl.copy()
                transient_tbl.path = all_tbl.savepath.transientcatalogpath
                if len(transient_tbl.data) > 0:
                    coord_first = SkyCoord(ra=transient_tbl.data['X_WORLD'],
                                        dec=transient_tbl.data['Y_WORLD'],
                                        unit = 'deg')
                    coord_second = SkyCoord(ra=reference_catalog.data['X_WORLD'],
                                            dec=reference_catalog.data['Y_WORLD'],
                                            unit = 'deg')
                    matched_first, matched_second, unmatched_first = self.helper.cross_match(coord_first, coord_second, subtract_seeing)
                    transient_tbl.data = transient_tbl.data[unmatched_first]
                
            # # Step 8: reverse subtraction (reference_img - target_img)
            # if reverse_subtraction:
            #     _, _, reverse_subframe_subtract_img, _, reverse_subframe_subtract_ivpmask = self.subtract(
            #         target_img = reprojected_reference_img,
            #         reference_img = reprojected_target_img,
            #         target_ivpmask = reprojected_reference_ivpmask,
            #         reference_ivpmask = reprojected_target_ivpmask,
            #         target_stamp = reprojected_target_stamp,
            #         id_ = i,
            #         save = save,
            #         verbose = verbose,
            #         visualize = visualize,
            #         nrx = 2,
            #         nry = 2,
            #         ko = 3,
            #         bgo = 3,
            #         nsx = 10,
            #         nsy = 10,
            #         r = 10,
            #         #ng = ng,
            #     )
            #     subtract_seeing = max(target_seeing, reference_seeing)
            #     tbl_third = self.aperphot.sex_photometry(
            #         target_img = reverse_subframe_subtract_img,
            #         target_bkg = None,  # No background for subtraction
            #         target_bkgrms = None,  # No background RMS for subtraction
            #         target_mask = reverse_subframe_subtract_ivpmask,
            #         sex_params = dict(SEEING_FWHM = subtract_seeing, BACK_TYPE = 'MANUAL'),
            #         detection_sigma = detection_sigma,
            #         aperture_diameter_arcsec = [6,9,12],
            #         saturation_level = 60000,
            #         save = save,
            #         verbose = verbose,
            #         visualize = visualize,
            #         save_fig = False
            #     )
            
            #     # Step 6: Filter the table for significant sources
            #     selected_tbl_third = self.select_transients(
            #         target_catalog = tbl_third,
            #         snr_lower = 3.0,
            #         fwhm_lower = transient_criteria['seeing_lower'],
            #         fwhm_upper = transient_criteria['seeing_upper'],
            #         flag_upper = 1,
            #         maskflag_upper = 1,
            #         class_star_lower = transient_criteria['classstar_lower'],
            #         elongation_upper = transient_criteria['elongation_upper'],
            #         flux_key = 'FLUX_AUTO',
            #         fluxerr_key = 'FLUXERR_AUTO',
            #         fwhm_key = 'FWHM_WORLD',
            #         flag_key = 'FLAGS',
            #         maskflag_key = 'IMAFLAGS_ISO',
            #         classstar_key = 'CLASS_STAR',
            #         elongation_key = 'ELONGATION',
            #         verbose = verbose,
            #         save = save,
            #         return_only_transient = True
            #     )

            #     # Remove 
            #     reverse_subframe_subtract_img.remove()
            #     reverse_subframe_subtract_ivpmask.remove()
                
            #     # Match the first and third photometry results
            #     if len(selected_tbl_third.data) > 0:
            #         coord_first = SkyCoord(ra=selected_tbl_first.data['X_WORLD'],
            #                             dec=selected_tbl_first.data['Y_WORLD'],
            #                             unit = 'deg')
            #         coord_third = SkyCoord(ra=selected_tbl_third.data['X_WORLD'],
            #                                 dec=selected_tbl_third.data['Y_WORLD'],
            #                                 unit = 'deg')
            #         matched_first, matched_third, unmatched_first = self.helper.cross_match(coord_first, coord_third, seeing_upper_criteria * 2)
            #         transient_tbl.data = transient_tbl.data[unmatched_first]
            #     else:
            #         self.print("No significant sources found in the third photometry step.", verbose)
            #         pass
            
            #     if len(transient_tbl.data) <= target_transient_number:
            #         transient_tbl = self.photcal.apply_zp(
            #             target_img = subframe_subtract_img,
            #             target_catalog = transient_tbl,
            #             save = True
            #         )
            
            if len(candidate_tbl.data) > 0:
                candidate_tbl = self.photcal.apply_zp(
                    target_img = subframe_subtract_img,
                    target_catalog = candidate_tbl,
                    save = True
                )   
            if len(transient_tbl.data) > 0:
                transient_tbl = self.photcal.apply_zp(
                    target_img = subframe_subtract_img,
                    target_catalog = transient_tbl,
                    save = True
                )
            
            if save_transient_figure:
                if transient_tbl is not None:
                    if len(transient_tbl.data) > 0:
                        all_transients = transient_tbl.data[:show_transient_numbers]
                        ra = all_transients['X_WORLD']
                        dec = all_transients['Y_WORLD']
                        idx = all_transients['NUMBER']
                        self.show_transient_positions(
                            science_img = subframe_target_img,
                            reference_img = subframe_reference_img,
                            subtracted_img = subframe_subtract_img,
                            x_list = ra,
                            y_list = dec,
                            idx_list = idx,
                            coord_type = 'coord',
                            zoom_radius_pixel = 30,
                            downsample = 1,
                            ncols = 3,
                            cmap = 'viridis',
                            scale = 'zscale',
                            subtitles = [f'Science', f'Reference', f'Subtracted'],
                            transient_type = 'transient',
                            save = save,
                            visualize = visualize
                        )
            if save_candidate_figure:
                if len(candidate_tbl.data) > 0:
                    all_candidates = candidate_tbl.data[:show_candidate_numbers]
                    ra = all_candidates['X_WORLD']
                    dec = all_candidates['Y_WORLD']
                    idx = all_candidates['NUMBER']
                    self.show_transient_positions(
                        science_img = subframe_target_img,
                        reference_img = subframe_reference_img,
                        subtracted_img = subframe_subtract_img,
                        x_list = ra,
                        y_list = dec,
                        idx_list = idx,
                        coord_type = 'coord',
                        zoom_radius_pixel = 30,
                        downsample = 1,
                        ncols = 3,
                        cmap = 'viridis',
                        scale = 'zscale',
                        subtitles = [f'Science', f'Reference', f'Subtracted'],
                        transient_type = 'candidate',
                        save = save,
                        visualize = visualize
                        )
                            
            final_subtraction_mask.write()
            all_catalogs.append(all_tbl)
            candidate_catalogs.append(candidate_tbl)
            transient_catalogs.append(transient_tbl)
            subframe_target_img.clear()
            subframe_reference_img.clear()
            subframe_subtract_img.clear()
            subframe_target_imglist.append(subframe_target_img)
            subframe_reference_imglist.append(subframe_reference_img)
            subframe_subtract_imglist.append(subframe_subtract_img)     
        return all_catalogs, candidate_catalogs, transient_catalogs, subframe_target_imglist, subframe_reference_imglist, subframe_subtract_imglist

    def select_transients(self,
                            target_catalog: Catalog,
                            snr_lower: float = 5.0,
                            fwhm_lower: float = 1.5,
                            fwhm_upper: float = 5.0,
                            flag_upper: int = 1,
                            maskflag_upper: int = 1,
                            class_star_lower: float = 0.9,
                            elongation_upper: float = 1.3,
                            flux_key: str = 'FLUX_AUTO',
                            fluxerr_key: str = 'FLUXERR_AUTO',
                            fwhm_key: str = 'FWHM_WORLD',
                            flag_key: str = 'FLAGS',
                            maskflag_key: str = 'NIMAFLAGS_ISO',
                            classstar_key: str = 'CLASS_STAR',
                            elongation_key: str = 'ELONGATION',
                            verbose: bool = True,
                            save: bool = True,
                            return_only_transient: bool = True):
        """
        Select valid sources from the catalog based on SNR, flags, class star, and elongation.
        
        Parameters
        ----------
        target_catalog : Catalog
            The catalog to select sources from.
        snr_lower : float, optional
            The lower limit for the SNR.    
        fwhm_lower : float, optional
            The lower limit for the FWHM.
        fwhm_upper : float, optional
            The upper limit for the FWHM.
        flag_upper : int, optional
            The upper limit for the flags.
        maskflag_upper : int, optional
            The upper limit for the mask flags.
        class_star_lower : float, optional
            The lower limit for the class star.
        elongation_upper : float, optional
            The upper limit for the elongation.
        flux_key : str, optional
            The key for the flux.
        fluxerr_key : str, optional
            The key for the flux error.
        fwhm_key : str, optional
            The key for the FWHM.
        flag_key : str, optional
            The key for the flags.
        maskflag_key : str, optional    
            The key for the mask flags.
        classstar_key : str, optional
            The key for the class star.
        elongation_key : str, optional
            The key for the elongation.
        verbose : bool, optional    
            Whether to print verbose output.
        save : bool, optional
            Whether to save the catalog.
        return_only_transient : bool, optional
            Whether to return only the transient catalog.

        Returns
        -------
        (target_catalog, candidate_catalog) : Tuple[Catalog, Catalog]
            The target catalog and the candidate catalog.
        """
        
        target_catalog_data = target_catalog.data
        target_catalog_data['SNR'] = target_catalog_data[flux_key] / target_catalog_data[fluxerr_key]
        if fwhm_key.upper() == 'FWHM_WORLD':
            target_catalog_data['SEEING'] = target_catalog_data['FWHM_WORLD'] * 3600
            fwhm_key = 'SEEING'
        
        snr_lower_idx = (target_catalog_data['SNR'] > snr_lower)
        fwhm_lower_idx = (target_catalog_data[fwhm_key] > fwhm_lower)
        fwhm_upper_idx = (target_catalog_data[fwhm_key] < fwhm_upper)
        flag_upper_idx = (target_catalog_data[flag_key] < flag_upper)
        maskflag_upper_idx = (target_catalog_data[maskflag_key] < maskflag_upper)
        classstar_lower_idx = (target_catalog_data[classstar_key] > class_star_lower)
        elongation_upper_idx = (target_catalog_data[elongation_key] < elongation_upper)
        all_idx = snr_lower_idx & fwhm_lower_idx & fwhm_upper_idx & flag_upper_idx & maskflag_upper_idx & classstar_lower_idx & elongation_upper_idx
        
        target_catalog_data['FLAG_SNR'] = snr_lower_idx
        target_catalog_data['FLAG_FWHM'] = fwhm_lower_idx & fwhm_upper_idx
        target_catalog_data['FLAG_MASK'] = maskflag_upper_idx
        target_catalog_data['FLAG_CLASS_STAR'] = classstar_lower_idx
        target_catalog_data['FLAG_ELONGATION'] = elongation_upper_idx
        target_catalog_data['FLAG_Transient'] = all_idx
        # Update the flags
        
        candidate_catalog = Catalog(path = target_catalog.savepath.candidatecatalogpath, catalog_type = 'transient', load = False)
        candidate_catalog.data = target_catalog_data[all_idx]
        target_catalog.data = target_catalog_data
        
        if verbose:
            print(f"Filtering sources based on criteria:")
            print(f"SNR > {snr_lower}: {np.sum(snr_lower_idx)}")
            print(f"{fwhm_key} > {fwhm_lower} and FWHM < {fwhm_upper}: {np.sum(fwhm_lower_idx & fwhm_upper_idx)}")
            print(f"{flag_key} < {flag_upper}: {np.sum(flag_upper_idx)}")
            print(f"{maskflag_key} < {maskflag_upper}: {np.sum(maskflag_upper_idx)}")
            print(f"{classstar_key} > {class_star_lower}: {np.sum(classstar_lower_idx)}")
            print(f"{elongation_key} < {elongation_upper}: {np.sum(elongation_upper_idx)}")
            print(f'Sources with all criteria met: {np.sum(all_idx)}')

        if save:
            target_catalog.write()
            candidate_catalog.write()
        
        return target_catalog, candidate_catalog
        
    def show_transient_positions(
        self,
        science_img: Union[ScienceImage, ReferenceImage],
        reference_img: Union[ScienceImage, ReferenceImage],
        subtracted_img: Union[ScienceImage, ReferenceImage],
        x_list: list,
        y_list: list,
        idx_list: list = None,
        coord_type='coord',
        zoom_radius_pixel=30,
        downsample=1,
        cmap='viridis',
        scale='zscale',
        ncols=2,
        subtitles: list = None,
        transient_type: str = 'transient',
        save: bool = True,
        visualize: bool = True
        ):
        """
        Show the transient positions on the science, reference, and subtracted images.
        
        Parameters
        ----------
        science_img : ScienceImage
            The science image.
        reference_img : ReferenceImage
            The reference image.
        subtracted_img : ScienceImage
            The subtracted image.
        x_list : list
            The list of x coordinates.
        y_list : list
            The list of y coordinates.
        idx_list : list, optional
            The list of indices.
        coord_type : str, optional
            The coordinate type. ['coord', 'pixel']
        zoom_radius_pixel : int, optional
            The zoom radius in pixels.
        downsample : int, optional
            The downsample factor.
        cmap : str, optional
            The colormap.
        scale : str, optional
            The scale. ['linear', 'log', 'sqrt', 'arcsinh']
        ncols : int, optional
            The number of columns.
        subtitles : list, optional
            The subtitles for the images.
        transient_type : str, optional
            The transient type. ['transient', 'candidate']
        save : bool, optional
            Whether to save the figure.
        visualize : bool, optional
            Whether to visualize the figure.
            
        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure.
        axes : list
            The axes.
        """
        import math
        import matplotlib.pyplot as plt
        import matplotlib.gridspec as gridspec
        import numpy as np
        from astropy.coordinates import SkyCoord
        import astropy.units as u

        if len(x_list) != len(y_list):
            raise ValueError("x_list and y_list must have the same length.")

        n_transients = len(x_list)
        n_rows = math.ceil(n_transients / ncols)

        height, width = science_img.data.shape
        aspect = width / height  # image aspect ratio

        # === Outer GridSpec: Controls spacing between triplets ===
        fig = plt.figure(figsize=(3 * ncols * aspect * 1.2, 2.5 * n_rows))
        outer_grid = gridspec.GridSpec(n_rows, ncols, figure=fig, wspace=0.3, hspace=0.6)  

        img_list = [science_img, reference_img, subtracted_img]
        subtitles = subtitles or ['Science', 'Reference', 'Residual']
        axes = []

        for idx, (x, y) in enumerate(zip(x_list, y_list)):
            row = idx // ncols
            col = idx % ncols
            idx_label = idx_list[idx] if idx_list is not None else None

            # Inner grid: 3 columns per triplet (science, ref, residual)
            inner_grid = gridspec.GridSpecFromSubplotSpec(
                1, 3, subplot_spec=outer_grid[row, col], wspace=0.15  # spacing within triplet
            )

            triplet_axes = [fig.add_subplot(inner_grid[0, j]) for j in range(3)]
            axes.extend(triplet_axes)

            # Plot images
            for j, (img, subtitle) in enumerate(zip(img_list, subtitles)):
                ax = triplet_axes[j]
                img.show_position(
                    x=x, y=y,
                    coord_type=coord_type,
                    zoom_radius_pixel=zoom_radius_pixel,
                    downsample=downsample,
                    cmap=cmap,
                    scale=scale,
                    ax=ax
                )

                # Formatting titles
                coord = SkyCoord(ra=x, dec=y, unit='deg')
                ra_str = coord.ra.to_string(unit=u.hourangle, sep='', pad=True, precision=2)   # HHMMSS.ss
                dec_str = coord.dec.to_string(sep='', alwayssign=True, pad=True, precision=1)  # ±DDMMSS.s
                objname = f'{transient_type}_J{ra_str}_{dec_str}'
                if idx_label is not None:
                    objname = f'[{idx_label:02d}] {objname}'

                ax.set_xticks([])
                ax.set_yticks([])
                if j != 1:  # Science & Residual only
                    ax.set_title(f"\n{subtitle}", fontsize=10, pad=6)
                else:
                    ax.set_title(f"{objname}\n{subtitle}", fontsize=10, pad=6)
        
        # Check scale
        for ax in fig.axes:
            for im in ax.get_images():
                vmin, vmax = im.get_clim()
                if not np.isfinite(vmin) or not np.isfinite(vmax) or vmin >= vmax:
                    print(f"[FIX] Invalid clim in axis {ax}, resetting.")
                    data = im.get_array()
                    vmin, vmax = np.nanmin(data), np.nanmax(data)
                    if vmin == vmax:  # flat image
                        vmin, vmax = vmin - 1, vmax + 1
                    im.set_clim(vmin, vmax)
        
        if save:
            fig.savefig(
                science_img.savepath.savedir / f'{science_img.savepath.savepath.name}.{transient_type}.png',
                bbox_inches='tight', dpi=300
            )
            
        if visualize:
            plt.show()
        else:
            plt.close()
        return fig, axes

    
    def get_referenceframe_from_image(self, 
                                      target_img: Union[ScienceImage],
                                      telname: str = None,
                                      max_obsdate: Union[str, float, Time] = None,
                                      sort_key: Union[str, List[str]] = ['fraction', 'depth'],
                                      overlap_threshold: float = 0.5,
                                      return_groups: bool = True,
                                      group_overlap_threshold: float = 0.8
                                      ):
        """
        Get the reference frame from the target image.
        
        Parameters
        ----------
        target_img : ScienceImage
            The target image to get the reference frame from.
        telname : str, optional
            The telescope name.
        max_obsdate : Union[str, float, Time], optional
            The maximum observation date.
        sort_key : Union[str, List[str]], optional
            The sort key.
        overlap_threshold : float, optional
            The overlap threshold.
        return_groups : bool, optional
            Whether to return the groups.
        group_overlap_threshold : float, optional
            The group overlap threshold.
            
        Returns
        -------
        reference_img : ReferenceImage
            The reference image.
        reference_frames : Table
            The metadata of the reference frames matched the criteria.
        """
        if max_obsdate is None:
            max_obsdate = target_img.obsdate
            print(f"No max_obsdate provided, using target image's obsdate instead ({max_obsdate}).")

        # Define fallback attempts
        attempt_configs = [
            {'seeing': target_img.seeing, 'depth': target_img.depth},
            {'seeing': target_img.seeing + 2, 'depth': target_img.depth},
            {'seeing': target_img.seeing + 2, 'depth': target_img.depth - 2},
        ]

        reference_frames = None
        for cfg in attempt_configs:
            reference_frames = self.get_referenceframe(
                observatory = target_img.observatory,
                telkey = target_img.telkey,
                filter_ = target_img.filter,
                ra = target_img.ra,
                dec = target_img.dec,
                ra_fov = target_img.fovx,
                dec_fov = target_img.fovy,
                telname = telname,
                max_obsdate = max_obsdate,
                seeing_limit = cfg['seeing'],
                depth_limit = cfg['depth'],
                return_groups = return_groups,
                overlap_threshold = overlap_threshold,
                group_overlap_threshold = group_overlap_threshold,
            )
            if reference_frames is not None and len(reference_frames) > 0:
                break  # success
                    
        if reference_frames is None:
            print("No reference frames found for the target image.")
            return None

        # Normalize to list
        if isinstance(sort_key, str):
            sort_key = [sort_key]

        # Determine sort direction for each key
        # Use descending (reverse=True) for 'depth' and 'fraction' by default
        reverse_map = {
            'depth': True,
            'fraction': True,
            'seeing': False,
            'obsdate': False,
        }
        reverse_flags = [reverse_map.get(k, False) for k in sort_key]

        # Apply sort
        reference_frames.sort(sort_key, reverse=reverse_flags)

        reference_img = ReferenceImage(reference_frames[0]['file'], telinfo=target_img.telinfo, load=True)
        return reference_img, reference_frames
    
    def get_referenceframe(self,
                           observatory: str,
                           telkey: str,
                           filter_: str,
                           ra: float,
                           dec: float,
                           ra_fov: float = 1.35,
                           dec_fov: float = 0.9,
                           telname: str = None,
                           max_obsdate: str = None,
                           seeing_limit: float = None,
                           depth_limit: float = None,
                           return_groups: bool = True,
                           overlap_threshold: float = 0.5,
                           group_overlap_threshold: float = 0.8
                           ):
        """
        Get the reference frame from the target image.
        
        Parameters
        ----------
        observatory : str
            The observatory name.
        telkey : str
            The telescope key.
        filter_ : str
            The filter name.
        ra : float
            The right ascension of the target image.
        dec : float
            The declination of the target image.
        ra_fov : float, optional
            The field of view in right ascension.
        dec_fov : float, optional
            The field of view in declination.
        telname : str, optional
            The telescope name.
        max_obsdate : str, optional
            The maximum observation date.
        seeing_limit : float, optional
            The seeing limit.
        depth_limit : float, optional
            The depth limit.
        return_groups : bool, optional
            Whether to return the groups.
        overlap_threshold : float, optional
            The overlap threshold.
        group_overlap_threshold : float, optional
            The group overlap threshold.
            
        Returns
        -------
        ref_table : Table
            The metadata of the reference frames matched the criteria.
        """
        
        # Load summary tables
        all_referenceframe_info = {}
        referenceframe_summary_path = Path(self.helper.config['REFDATA_DIR']) / 'summary.ascii_fixed_width'

        if referenceframe_summary_path.exists():
            tbl = ascii.read(referenceframe_summary_path, format='fixed_width')
            all_referenceframe_tbl = tbl
        else:
            all_referenceframe_tbl = Table()
                
        if len(all_referenceframe_tbl) == 0:
            raise FileNotFoundError("No calibration frame metadata found.")

        # Basic filtering
        mask = np.ones(len(all_referenceframe_tbl), dtype=bool)

        # Apply filters only if not None
        mask &= all_referenceframe_tbl['observatory'] == observatory
        mask &= all_referenceframe_tbl['telkey'] == telkey
        mask &= all_referenceframe_tbl['filtername'] == filter_
        if telname is not None:
            mask &=  np.array([telname in str(row) for row in all_referenceframe_tbl['telname']])
        if max_obsdate is not None:
            obsdate_target = self.helper.flexible_time_parser(max_obsdate)
            obs_times = Time(all_referenceframe_tbl['obsdate'], format='isot', scale='utc')
            mask &= obs_times < obsdate_target
        if seeing_limit is not None:
            mask &= all_referenceframe_tbl['seeing'] <= seeing_limit
        if depth_limit is not None:
            mask &= all_referenceframe_tbl['depth'] >= depth_limit
        
        filtered_tbl = all_referenceframe_tbl[mask]
        
        if len(filtered_tbl) == 0:
            try:
                print(f"No reference frames matched the filtering criteria. [Depth > %.1f, Seeing < %.1f, Obsdate <= %s]" %(depth_limit, seeing_limit, obsdate_target))
            except:
                print(f"No reference frames matched the filtering criteria.Obsdate <= %s" % max_obsdate)
            return None
        else:
            pass
        
        # Geometry filtering using RA, Dec, FOV
        target_poly = Polygon([
            (ra - ra_fov / 2, dec - dec_fov / 2),
            (ra + ra_fov / 2, dec - dec_fov / 2),
            (ra + ra_fov / 2, dec + dec_fov / 2),
            (ra - ra_fov / 2, dec + dec_fov / 2)
        ])
        target_area = target_poly.area

        # Geometry filtering: keep only intersecting reference frames
        matched_rows = []
        fractions = []

        for row in filtered_tbl:
            ref_poly = Polygon([
                (row['ra'] - row['fov_ra'] / 2, row['dec'] - row['fov_dec'] / 2),
                (row['ra'] + row['fov_ra'] / 2, row['dec'] - row['fov_dec'] / 2),
                (row['ra'] + row['fov_ra'] / 2, row['dec'] + row['fov_dec'] / 2),
                (row['ra'] - row['fov_ra'] / 2, row['dec'] + row['fov_dec'] / 2)
            ])

            if ref_poly.intersects(target_poly):
                matched_rows.append(row)
                inter_area = target_poly.intersection(ref_poly).area
                frac = inter_area / target_area if target_area > 0 else 0.0
                fractions.append(frac)

        if len(matched_rows) == 0:
            print(f"No reference frames found overlapping RA={ra}, Dec={dec} with FOV=({ra_fov}, {dec_fov})")
            return None

        # Build final table with overlap fractions
        ref_table = Table(rows=matched_rows, names=filtered_tbl.colnames)
        ref_table['fraction'] = fractions
        ref_table = ref_table[ref_table['fraction'] > overlap_threshold]

        # Optional: assign overlap-based group IDs
        if return_groups:
            def assign_groups(ref_table: Table, overlap_threshold: float = 0.8) -> Table:
                n = len(ref_table)
                polygons = []
                for row in ref_table:
                    poly = Polygon([
                        (row['ra'] - row['fov_ra'] / 2, row['dec'] - row['fov_dec'] / 2),
                        (row['ra'] + row['fov_ra'] / 2, row['dec'] - row['fov_dec'] / 2),
                        (row['ra'] + row['fov_ra'] / 2, row['dec'] + row['fov_dec'] / 2),
                        (row['ra'] - row['fov_ra'] / 2, row['dec'] + row['fov_dec'] / 2),
                    ])
                    polygons.append(poly)

                adjacency = [set() for _ in range(n)]
                for i in range(n):
                    for j in range(i + 1, n):
                        inter_area = polygons[i].intersection(polygons[j]).area
                        min_area = min(polygons[i].area, polygons[j].area)
                        if min_area > 0 and (inter_area / min_area) >= overlap_threshold:
                            adjacency[i].add(j)
                            adjacency[j].add(i)

                visited = [False] * n
                group_ids = [-1] * n
                group = 0
                for i in range(n):
                    if not visited[i]:
                        queue = [i]
                        while queue:
                            current = queue.pop()
                            if not visited[current]:
                                visited[current] = True
                                group_ids[current] = group
                                queue.extend(adjacency[current])
                        group += 1

                ref_table['group'] = group_ids
                return ref_table

            ref_table = assign_groups(ref_table, overlap_threshold=group_overlap_threshold)

        return ref_table
    
    def query_referenceframe_from_image(self,
                                        target_img: Union[ScienceImage],
                                        catalog_key: str = 'SkyMapper/SMSS4/g',
                                        ):
        """
        Query the reference frame from the target image using HIPS2FITS.
        
        Parameters
        ----------
        target_img : ScienceImage
            The target image to query the reference frame from.
        catalog_key : str, optional 
            The catalog key to query the reference frame from.
            
        Returns
        -------
        reference_img : ReferenceImage
            The reference image.
        """
        imagequerier = ImageQuerier(catalog_key = catalog_key)
        reference_path = target_img.savedir / f'{target_img.objname}_ref.fits'
        reference_img = imagequerier.query(
            width = int(target_img.naxis1 * 1.2),
            height = int(target_img.naxis2 * 1.2),
            ra = target_img.ra,
            dec = target_img.dec,
            pixelscale = np.mean(target_img.pixelscale),
            telinfo = target_img.telinfo,
            save_path = reference_path,
            objname = target_img.objname,
        )      
        return reference_img
         
    def select_reference_image(self, 
                               target_imglist: Union[List[ScienceImage], List[ReferenceImage]],
                               max_obsdate: Union[Time, str, float] = None,
                               seeing_key: str = 'SEEING',
                               depth_key: str = 'UL5_5',
                               ellipticity_key: str = 'ELLIP',
                               obsdate_key: str = 'DATE-OBS',
                               weight_ellipticity: float = 2.0,
                               weight_seeing: float = 1.0,
                               weight_depth: float = 1.5,
                               max_numbers: int = 1):
        """
        Select the reference image from the target image list.
        
        Parameters
        ----------
        target_imglist : List[ScienceImage]
            The list of target images.
        max_obsdate : Union[Time, str, float], optional
            The maximum observation date.
        seeing_key : str, optional
            The seeing key.
        depth_key : str, optional
            The depth key.
        ellipticity_key : str, optional
            The ellipticity key.
        obsdate_key : str, optional
            The observation date key.
        weight_ellipticity : float, optional
            The weight for the ellipticity.
        weight_seeing : float, optional
            The weight for the seeing.
        weight_depth : float, optional
            The weight for the depth.
        max_numbers : int, optional
            The maximum number of images to select.
            
        Returns
        -------
        best_image : ScienceImage
            The best image.
        """
        
        seeinglist = []
        depthlist = []
        ellipticitylist = []
        obsdatelist = []
        for target_img in tqdm(target_imglist, desc = 'Querying reference images...'):
            seeinglist.append(target_img.header.get(seeing_key, None))
            depthlist.append(target_img.header.get(depth_key, None))
            ellipticitylist.append(target_img.header.get(ellipticity_key, None))
            obsdatelist.append(target_img.header.get(obsdate_key, None))
        
        try:
            obsdate_time = Time(obsdatelist)
            max_obs_time = self.helper.flexible_time_parser(max_obsdate) if max_obsdate is not None else Time.now()
        except Exception as e:
            raise ValueError(f"Invalid date format: {e}")          
        # Mask for images before max_obsdate
        valid_obs_mask = obsdate_time < max_obs_time
        
        # Also apply validity mask for seeing, depth, ellipticity
        seeinglist = np.array(seeinglist, dtype=float)
        depthlist = np.array(depthlist, dtype=float)
        ellipticitylist = np.array(ellipticitylist, dtype=float)
        valid_value_mask = (~np.isnan(seeinglist)) & (~np.isnan(depthlist)) & (~np.isnan(ellipticitylist))
        combined_mask = valid_obs_mask & valid_value_mask
          
        # Apply final mask
        ell = np.array(ellipticitylist)[combined_mask]
        see = np.array(seeinglist)[combined_mask]
        dep = np.array(depthlist)[combined_mask]
        filtered_imgs = np.array(target_imglist)[combined_mask]
        filtered_obsd = np.array(obsdate_time)[combined_mask]

        # Normalize
        scaler = MinMaxScaler()
        ell_norm = scaler.fit_transform(ell.reshape(-1, 1)).flatten()
        see_norm = scaler.fit_transform(see.reshape(-1, 1)).flatten()
        dep_norm = scaler.fit_transform(dep.reshape(-1, 1)).flatten()

        # Compute combined score
        # You can adjust weights if needed
        score = (1 - ell_norm) * weight_ellipticity + (1 - see_norm) * weight_seeing + dep_norm * weight_depth

        # Rank and select best
        sorted_idx = np.argsort(score)[::-1]  # descending
        best_images = filtered_imgs[sorted_idx]

        # Top N or just best
        best_image = best_images[0]
        
        # Data for plotting
        x = np.array(seeinglist)
        y = np.array(depthlist)
        c = np.array(ellipticitylist)
        x_valid = x[combined_mask]
        y_valid = y[combined_mask]
        c_valid = c[combined_mask]
        best_idx = sorted_idx[0]
        best_x = see[best_idx]
        best_y = dep[best_idx]
        best_c = ell[best_idx]
        selected_idx = sorted_idx[:max_numbers]
        selected_x = see[selected_idx]
        selected_y = dep[selected_idx]
        selected_c = ell[selected_idx]
        marker_sizes = np.where(obsdate_time < max_obs_time, 50, 10)
        marker_alphas = np.where(obsdate_time < max_obs_time, 0.8, 0.2)

        # Calculate percentiles (90%, 75%, and 50%)
        p90_x, p75_x, p50_x, p25_x, p10_x = np.percentile(x_valid, [10, 25, 50, 75, 90])
        p90_y, p75_y, p50_y, p25_y, p10_y = np.percentile(y_valid, [90, 75, 50, 25, 10])

        # Calculate the number of images for each percentile
        num_images_p90 = np.sum((x_valid <= p90_x) & (y_valid >= p90_y))  # Number of images below or equal to the 10th percentile
        num_images_p75 = np.sum((x_valid <= p75_x) & (y_valid >= p75_y))  # Number of images below or equal to the 25th percentile
        num_images_p50 = np.sum((x_valid <= p50_x) & (y_valid >= p50_y))  # Number of images below or equal to the 50th percentile
        num_images_p25 = np.sum((x_valid <= p25_x) & (y_valid >= p25_y))  # Number of images below or equal to the 75th percentile

        # Create figure with GridSpec layout
        fig = plt.figure(figsize=(6, 6), dpi=300)
        gs = GridSpec(4, 4, fig)

        # Create scatter plot
        ax_main = fig.add_subplot(gs[1:, :-1])
        sc = ax_main.scatter(x[valid_value_mask], y[valid_value_mask],
                            c=c[valid_value_mask],
                            s=marker_sizes[valid_value_mask],
                            alpha=marker_alphas[valid_value_mask],
                            cmap='viridis', edgecolors='k', linewidths=0.5,
                            label = 'All images')        
        ax_main.scatter(0,0, s = 10, alpha = 0.2, label = 'Out of date range')
        cbar = fig.colorbar(sc, ax=ax_main, pad=0.01)
        cbar.set_label('Ellipticity')
        ax_main.axvline(p90_x, color='r', linestyle='--')
        ax_main.axvline(p75_x, color='b', linestyle='--')
        ax_main.axvline(p50_x, color='g', linestyle='--')
        ax_main.axvline(p25_x, color='k', linestyle='--')
        ax_main.axhline(p90_y, color='r', linestyle='--')
        ax_main.axhline(p75_y, color='b', linestyle='--')
        ax_main.axhline(p50_y, color='g', linestyle='--')
        ax_main.axhline(p25_y, color='k', linestyle='--')
        ax_main.set_xlim(p90_x - 0.5, p10_x + 0.5)
        ax_main.set_ylim(p10_y - 1, p90_y + 1)
        ax_main.set_xlabel('Seeing [arcsec]')
        ax_main.set_ylabel('Depth [AB]')
        ax_main.scatter(selected_x, selected_y, marker='*', s=200, c='red', edgecolors='black', label='Selected')
        ax_main.scatter(best_x, best_y, marker='*', s=200, c='red', edgecolors='black')
        ax_main.text(best_x, best_y + 0.3,
                    f"Best\nSeeing = {best_x:.2f} arcsec\nDepth = {best_y:.2f} AB\nEllipticity = {best_c:.2f}",
                    color='red', fontsize=8, ha='center', va='bottom',
                    bbox=dict(facecolor='white', edgecolor='red', boxstyle='round,pad=0.3'))
        ax_main.legend(loc='upper right', fontsize=8, frameon=True)


        # Create top histogram
        ax_histx = fig.add_subplot(gs[0, :-1], sharex=ax_main)
        ax_histx.hist(x_valid, bins=30, color='black', edgecolor='black', alpha=0.7)
        ax_histx.spines['top'].set_visible(False)  # Hide top spine
        ax_histx.spines['right'].set_visible(False)  # Hide right spine

        # Create right histogram
        ax_histy = fig.add_subplot(gs[1:, -1], sharey=ax_main)
        ax_histy.hist(y_valid, bins=30, color='black', edgecolor='black', alpha=0.7, orientation='horizontal')
        ax_histy.spines['top'].set_visible(False)  # Hide top spine
        ax_histy.spines['right'].set_visible(False)  # Hide right spine

        # Set limits for histograms to fit within the black box
        ax_histx.set_xlim(ax_main.get_xlim())
        ax_histy.set_ylim(ax_main.get_ylim())

        # Plot vertical regions for percentiles in histograms
        ax_histx.axvline(p90_x, color='r', linestyle='--', label='90%')
        ax_histx.axvline(p75_x, color='b', linestyle='--', label='75%')
        ax_histx.axvline(p50_x, color='g', linestyle='--', label='50%')
        ax_histx.axvline(p25_x, color='k', linestyle='--', label='25%')

        ax_histy.axhline(p90_y, color='r', linestyle='--', label='90%')
        ax_histy.axhline(p75_y, color='b', linestyle='--', label='75%')
        ax_histy.axhline(p50_y, color='g', linestyle='--', label='50%')
        ax_histy.axhline(p25_y, color='k', linestyle='--', label='25%')

        # Add text annotation in the upper right region of the scatter plot
        text = f'Percentile (# of images, Seeing, Depth):\n'
        text += f'90% ({num_images_p90}, {p90_x:.2f}, {p90_y:.2f})\n'
        text += f'75% ({num_images_p75}, {p75_x:.2f}, {p75_y:.2f})\n'
        text += f'50% ({num_images_p50}, {p50_x:.2f}, {p50_y:.2f})\n'
        text += f'25% ({num_images_p25}, {p25_x:.2f}, {p25_y:.2f})'
        ax_main.text(0.5, 0.15, text,
                    ha='center', va='center',
                    transform=ax_main.transAxes,
                    bbox=dict(facecolor='white', edgecolor='black', boxstyle='round,pad=0.5'))
        
        from matplotlib.lines import Line2D

        dashed_lines = [
            Line2D([0], [0], color='red', linestyle='--', label='90%'),
            Line2D([0], [0], color='blue', linestyle='--', label='75%'),
            Line2D([0], [0], color='green', linestyle='--', label='50%'),
            Line2D([0], [0], color='black', linestyle='--', label='25%')
        ]

        fig.legend(handles=dashed_lines,
                loc='upper right',
                bbox_to_anchor=(0.95, 0.95),
                fontsize=10, frameon=True)
        plt.tight_layout()
        plt.show()
        
        selected_images = []
        for img in tqdm(best_images[:max_numbers], desc='Loading selected images...'):
            ref_img = ReferenceImage(img.path, img.telinfo, load=True)
            ref_img.load()
            ref_img.path = ref_img.savepath.savepath
            selected_images.append(ref_img)
        return selected_images

    def _reproject_to_target(self,
                             reference_img: ReferenceImage,
                             target_img: ScienceImage,
                             save: bool = True,
                             verbose: bool = True):
        """
        Reproject the reference image to the target image's central WCS
        """
        reference_img.savedir = target_img.savepath.savedir
        center_target = target_img.center
        reprojected_reference, _, reprojected_reference_ivpmask = self.projection.reproject(
            target_img = reference_img,
            swarp_params = None,
            resample_type = 'LANCZOS3',
            center_ra = center_target['ra'],
            center_dec = center_target['dec'],
            x_size = target_img.naxis1,
            y_size = target_img.naxis2,
            pixelscale = target_img.pixelscale.mean(),
            verbose = verbose,
            overwrite = False,
            save = save,
            return_ivpmask = True
        )
        
        if not self.helper.is_wcs_equal(reprojected_reference.wcs, target_img.wcs):
            self.helper.print(f"Warning: target_img is not reprojected (not aligned to the North). Run Projection().reproject", verbose)
        return reprojected_reference, reprojected_reference_ivpmask

    def _prepare_subtract_region(self,
                                 target_img: ScienceImage,
                                 reference_img: ReferenceImage,
                                 target_ivpmask: Mask = None,
                                 reference_ivpmask: Mask = None,
                                 target_stamp: str = None,
                                 id_: int = 0,
                                 save: bool = True,
                                 verbose: bool = True,
                                 visualize: bool = True):
        """
        target_img: should be reprojected
        reference_img: should be reprojected
        So, both images should have the same WCS.
        target_ivpmask: Mask for the target image, if None, will be created
        reference_ivpmask: Mask for the reference image, if None, will be created
        This will create the subtracted region by trimming both images to the overlapping region.
        fullframe (of target_img) subtraction region also will be returned 
        """
        
        # If wcs is not equal, reproject target to reference
        if not self.helper.is_wcs_equal(reference_img.wcs, target_img.wcs):
            raise RuntimeError("Target and reference images have different WCS. Please reproject the target image to match the reference image WCS.")
        else:
            if target_ivpmask is None:
                target_ivpmask = self.masking.mask_invalidpixel(
                    target_img = target_img,
                    target_mask = None,
                    save = False,
                    verbose = verbose,
                    visualize = False,
                    save_fig = False
                )
            if reference_ivpmask is None:
                reference_ivpmask = self.masking.mask_invalidpixel(
                    target_img = reference_img,
                    target_mask = None,
                    save = False,
                    verbose = verbose,
                    visualize = False,
                    save_fig = False
                )
            
        # All data and mask are ready
        # Create a new mask that combines the invalid pixel masks
        fullframe_subtract_mask = target_ivpmask.copy()
        fullframe_subtract_mask.masktype = 'subtraction'
        fullframe_subtract_mask.path = target_img.savepath.submaskpath
        fullframe_subtract_mask.combine_mask(reference_ivpmask.data, operation='or')
        fullframe_subtract_mask.data = 1 - fullframe_subtract_mask.data
        
        valid_mask = fullframe_subtract_mask.data == 1
        y_valid, x_valid = np.where(valid_mask)
        y_min, y_max = np.min(y_valid), np.max(y_valid) + 1
        x_min, x_max = np.min(x_valid), np.max(x_valid) + 1
        shape = (y_max - y_min, x_max - x_min)
        position = (x_min + shape[1] // 2, y_min + shape[0] // 2)  # (x_center, y_center)
        
        cutout_target = Cutout2D(data=target_img.data, position=position, size=shape, wcs=target_img.wcs)
        cutout_reference = Cutout2D(data=reference_img.data, position=position, size=shape, wcs=reference_img.wcs)

        # Update image and WCS
        subframe_target_img = target_img.copy()
        subframe_target_img.path = target_img.savepath.savedir / (target_img.savepath.savepath.stem + f'_subframe_{id_}' + target_img.savepath.savepath.suffix)
        subframe_target_img.data = cutout_target.data
        subframe_target_img.header.update(cutout_target.wcs.to_header())

        subframe_reference_img = reference_img.copy()
        subframe_reference_img.path = subframe_reference_img.savepath.savedir / (subframe_reference_img.savepath.savepath.stem + f'_subframe_{id_}' + subframe_reference_img.savepath.savepath.suffix)
        subframe_reference_img.savedir = subframe_target_img.savedir # Change the savedir to the same as target_img
        subframe_reference_img.data = cutout_reference.data
        subframe_reference_img.header.update(cutout_reference.wcs.to_header())
        
        subframe_target_ivpmask = target_ivpmask.copy()
        subframe_target_ivpmask.path = subframe_target_img.savepath.invalidmaskpath#target_ivpmask.savepath.savedir / (target_ivpmask.savepath.savepath.stem + f'_subframe_{id_}' + target_ivpmask.savepath.savepath.suffix)
        subframe_target_ivpmask.data = target_ivpmask.data[y_min:y_max, x_min:x_max]
        subframe_target_ivpmask.header.update(cutout_target.wcs.to_header())
        
        subframe_reference_ivpmask = reference_ivpmask.copy()
        subframe_reference_ivpmask.path = subframe_reference_img.savepath.invalidmaskpath#reference_ivpmask.savepath.savedir / (reference_ivpmask.savepath.savepath.stem + f'_subframe_{id_}' + reference_ivpmask.savepath.savepath.suffix)
        subframe_reference_ivpmask.data = reference_ivpmask.data[y_min:y_max, x_min:x_max]
        subframe_reference_ivpmask.header.update(cutout_reference.wcs.to_header())
        
        subframe_subtract_mask = subframe_target_ivpmask.copy()
        subframe_subtract_mask.masktype = 'subtraction'
        subframe_subtract_mask.path = 'None'
        subframe_subtract_mask.combine_mask(subframe_reference_ivpmask.data, operation='or')
        
        # If target_stamp is provided, trim the subframe_target_img to the target_stamp
        subframe_target_stamp_path = None
        if target_stamp is not None:
            target_stamp = Path(target_stamp)
            if not target_stamp.exists():
                self.helper.print(f"Target stamp file {target_stamp} does not exist.", verbose)
            else:
                stamp_tbl = Table.read(target_stamp, format='ascii')
                x_key = stamp_tbl.colnames[0]
                y_key = stamp_tbl.colnames[1]
                x_full = np.array(stamp_tbl[x_key])
                y_full = np.array(stamp_tbl[y_key])

                # Filter for sources within the cutout region
                in_cutout = (
                    (x_full >= x_min) & (x_full < x_max) &
                    (y_full >= y_min) & (y_full < y_max)
                )

                # Shift coordinates to subframe system
                x_sub = x_full[in_cutout] - x_min
                y_sub = y_full[in_cutout] - y_min
                
                subframe_target_stamp = Table()
                subframe_target_stamp[x_key] = x_sub
                subframe_target_stamp[y_key] = y_sub
                subframe_target_stamp_path = target_stamp.parent / (target_stamp.stem + f'_subframe_{id_}' + target_stamp.suffix)
                subframe_target_stamp.write(subframe_target_stamp_path, format='ascii', overwrite=True)
        
        # Fill nan value to 0
        subframe_target_img.data += 10* subframe_target_img.header['SKYSIG']
        subframe_reference_img.data += 10* subframe_reference_img.header['SKYSIG']
        subframe_target_img.data[np.isnan(subframe_target_img.data)] = 0
        subframe_reference_img.data[np.isnan(subframe_reference_img.data)] = 0  

        if visualize:
            subframe_target_img.show()
            subframe_reference_img.show()
        
        if save:
            subframe_target_img.write()
            subframe_reference_img.write()
            subframe_target_ivpmask.write() 
            subframe_reference_ivpmask.write()
        
        return subframe_target_img, subframe_reference_img, subframe_target_ivpmask, subframe_reference_ivpmask, fullframe_subtract_mask, subframe_subtract_mask, subframe_target_stamp_path




# %%
if __name__ == "__main__":
    from ezphot.utils import DataBrowser
    self = TIPSubtraction()
#%%
if __name__ == "__main__":
    databrowser = DataBrowser('scidata')
    tile_id = 'T08262'
    databrowser.objname = tile_id
    databrowser.filter = 'r'
    stacked_imgset = databrowser.search(
        pattern = 'calib*202507*com.fits',
        return_type = 'science'
    )        
    target_imglist = stacked_imgset.target_images
    target_img = target_imglist[0]
    target_bkg = target_img.bkgmap
    reference_img = self.get_referenceframe_from_image(target_img)[0]
    reference_imglist = [reference_img]
    target_bkg: Background = None
    detection_sigma: float = 5
    aperture_diameter_arcsec: float = [5,7,10]
    aperture_diameter_seeing: float = [3.5,4.5]
    kron_factor = 1.5 
    
    target_transient_number: int = 5
    reject_variable_sources: bool = True
    negative_detection: bool = True
    reverse_subtraction: bool = False
    
    save: bool = True
    verbose: bool = True
    visualize: bool = True
    show_transient_numbers: int = 10
    nrx: int = 1
    nry: int = 1
    nsx: int = 10
    nsy: int = 10
    ko: int = 3
    bgo: int = 1
    r: int = 10
    hotpants_params = dict()
    # result = self.find_transients(
    #     target_img = target_img,
    #     reference_imglist = reference_imglist,
    #     target_bkg = None,
    #     detection_sigma = 5,
    #     aperture_diameter_arcsec = [5,7,10],
    #     aperture_diameter_seeing = [3.5,4.5],
    #     target_transient_number = 5,
    #     reject_variable_sources = True,
    #     negative_detection = True,
    #     reverse_subtraction = False,
    #     visualize = False,
    # )


# %%



#%%
from typing import Union, Optional
from pathlib import Path
import matplotlib.pyplot as plt
from astropy.stats import SigmaClip, sigma_clipped_stats
from photutils.segmentation import detect_threshold, detect_sources, SourceCatalog
from photutils.utils import circular_footprint
from mpl_toolkits.axes_grid1 import make_axes_locatable
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
import numpy as np
from photutils.aperture import CircularAperture
import astroscrappy as cr
from skimage.draw import disk
from tqdm import tqdm

from ezphot.imageobjects import Mask
from ezphot.imageobjects import ScienceImage, ReferenceImage, CalibrationImage
from ezphot.helper import Helper
#%%
class MaskGenerator():
    """
    Method class to generate masks for astronomical images.
    
    This class provides methods 
    
    1. Generation of invalid pixel mask with NaN and large connected regions of zero value.
    
    2. Generation of source mask with SExtractor-like detection.
    
    3. Generation of circular mask for given position and radius.
    
    4. Generation of cosmic ray mask with Astroscrappy.
    """
    
    def __init__(self):
        self.helper = Helper()
                
    def mask_invalidpixel(self,
                         # Input parameters
                          target_img: Union[ScienceImage, ReferenceImage, CalibrationImage],
                          target_mask: Optional[Mask] = None,
                          threshold_invalid_connection: int = 100000,
                          # Others
                          save: bool = False,
                          verbose: bool = True,
                          visualize: bool = True,
                          save_fig: bool = False):
        """
        Generate invalid pixel mask.
        
        Parameters
        ----------
        target_img : ScienceImage or ReferenceImage or CalibrationImage
            The target image to generate the invalid pixel mask from.
        target_mask : Mask, optional
            The mask to use for the invalid pixel mask generation.
        threshold_invalid_connection : int, optional
            The threshold for the invalid pixel mask generation.
        save : bool, optional
            Whether to save the invalid pixel mask.
        verbose : bool, optional
            Whether to print verbose output.
        visualize : bool, optional
            Whether to visualize the invalid pixel mask.
        save_fig : bool, optional
            Whether to save the invalid pixel mask as a figure.
            
        Returns
        -------
        target_mask : Mask
            The invalid pixel mask.
        """
        
        import numpy as np
        from scipy.ndimage import label

        if target_mask is None:
            target_mask = Mask(target_img.savepath.invalidmaskpath, masktype='invalid', load=False)
            target_mask.remove()
        else:
            self.helper.print("External mask is loaded.", verbose)

        image_data = target_img.data

        # Mask NaNs
        nan_mask = np.isnan(image_data)
        if np.any(nan_mask):
            self.helper.print(f"Masked {np.sum(nan_mask)} NaN pixels.", verbose)

        # Mask large connected regions of 0s
        zero_mask = (np.abs(image_data) == 0)
        labeled_array, num_features = label(zero_mask)
        self.helper.print(f"Found {num_features} connected regions", verbose)

        # Efficient region size filtering using np.bincount
        label_sizes = np.bincount(labeled_array.ravel())
        large_labels = np.where(label_sizes > threshold_invalid_connection)[0]
        large_labels = large_labels[large_labels != 0]  # exclude background

        large_zero_mask = np.isin(labeled_array, large_labels)
        self.helper.print(f"{len(large_labels)} regions larger than {threshold_invalid_connection} pixels", verbose)

        # Combine all invalid pixel masks
        invalidmask = nan_mask | large_zero_mask

        mask_previous = target_mask.data
        target_mask.combine_mask(invalidmask, 'or')
        target_mask.header = target_img.header

        # Update header/status
        update_header_kwargs = dict(
            TGTPATH=str(target_img.path),
            MASKTYPE='InvalidPixel'
        )
        target_mask.header.update(update_header_kwargs)

        event_details = dict(
            nan_masked=str(np.sum(nan_mask)),
            num_zero_regions=str(num_features),
            threshold_invalid_connection=str(threshold_invalid_connection),
            zero_masked=str(np.sum(large_zero_mask))
        )
        target_mask.add_status("invalid_mask", **event_details)

        if save:
            target_mask.write()

        # Visualize the mask
        if visualize or save_fig:
            save_path = None
            if save_fig:
                save_path = str(target_mask.savepath.savepath) + '.png'
            self._visualize(
                target_img=target_img,
                final_mask=target_mask,
                previous_mask=mask_previous,
                save_path=save_path,
                show=visualize
            )

        return target_mask

    def mask_sources(self, 
                     # Input parameters
                     target_img: Union[ScienceImage, ReferenceImage, CalibrationImage],
                     target_mask: Optional[Mask] = None,
                     sigma: float = 5.0, 
                     mask_radius_factor: float = 3,
                     saturation_level: float = 50000,
                     
                     # Others
                     save: bool = False,
                     verbose: bool = True,
                     visualize: bool = True,
                     save_fig: bool = False,
                     ): 
        """
        Generate source mask.
        
        Parameters
        ----------
        target_img : ScienceImage or ReferenceImage or CalibrationImage
            The target image to generate the source mask from.
        target_mask : Mask, optional
            The mask to use for the source mask generation.
        sigma : float, optional
            The sigma for the source mask generation.
        mask_radius_factor : float, optional
            The mask radius factor for the source mask generation.
        saturation_level : float, optional
            The saturation level for the source mask generation.
        save : bool, optional
            Whether to save the source mask.
        verbose : bool, optional
            Whether to print verbose output.
        visualize : bool, optional
            Whether to visualize the source mask.
        save_fig : bool, optional
            Whether to save the source mask as a figure.
            
        Returns
        -------
        target_mask : Mask
            The source mask.
        """
        if target_mask is None:
            target_mask = Mask(target_img.savepath.srcmaskpath, masktype = 'source', load=False)
        else:
            self.helper.print("External mask is loaded.", verbose)
        self.helper.print(f"Masking source... [sigma = {sigma}, mask_radius_factor = {mask_radius_factor}]", verbose)
        npixels = self.helper.load_config(target_img.config['SEX_CONFIG'])['DETECT_MINAREA']
        image_data, image_header = target_img.data, target_img.header
        sigma_clip = SigmaClip(sigma=sigma)
        threshold = detect_threshold(data = image_data, nsigma=sigma/np.sqrt(npixels), mask = target_mask.data, sigma_clip=sigma_clip)
        segment_img = detect_sources(image_data, threshold, npixels=npixels) 
        
        if segment_img:
            S = SourceCatalog(image_data, segment_img)
            props = S.to_table()
            new_mask = np.zeros(image_data.shape, dtype=bool)
            
            # Split props into saturated and non-saturated sources
            non_sat_rows = props[props['max_value'] <= saturation_level]
            sat_rows = props[props['max_value'] > saturation_level]

            self.helper.print(f"{len(non_sat_rows)} non-saturated, {len(sat_rows)} saturated sources", verbose)

            # Mask non-saturated sources (scaled by area)
            for row in non_sat_rows:
                area = row['area'].value
                y, x = row['ycentroid'], row['xcentroid']
                radius = mask_radius_factor * np.sqrt(area / np.pi)
                rr, cc = disk((y, x), radius, shape=image_data.shape)
                new_mask[rr, cc] = True

            # Mask saturated sources (fixed large radius)
            for row in sat_rows:
                area = row['area'].value
                y, x = row['ycentroid'], row['xcentroid']
                saturated_radius = mask_radius_factor * 2 * np.sqrt(area / np.pi)
                rr, cc = disk((y, x), saturated_radius, shape=image_data.shape)
                new_mask[rr, cc] = True

            mask_previous = target_mask.data
            target_mask.combine_mask(new_mask, 'or')
            target_mask.header = target_img.header
            
            # Update header/status
            update_header_kwargs = dict(
                TGTPATH = str(target_img.path),
                )
            
            current_masktype = target_mask.info.MASKTYPE
            if 'MASKTYPE' not in target_mask.header:
                update_header_kwargs['MASKTYPE'] = "Source"
            else:
                if "Source" not in current_masktype:
                    update_header_kwargs['MASKTYPE'] = f"{current_masktype},Source"
            
            ## Update attempt
            if 'MASKATMP' not in target_mask.info.to_dict():
                update_header_kwargs['MASKATMP'] = 1
            else:
                update_header_kwargs['MASKATMP'] = int(target_mask.info.MASKATMP) + 1
            target_mask.header.update(update_header_kwargs)
                
            ## Update status          
            event_details = dict(sigma = sigma, mask_radius_factor = mask_radius_factor, num_mask = segment_img.nlabels)
            target_mask.add_status("source_mask", **event_details)
            self.helper.print(f"{segment_img.nlabels} sources masked.", verbose)
            
            # Save mask
            if save:
                target_mask.write()
            
            # Visualize
            if visualize or save_fig:
                save_path = None
                if save_fig:
                    save_path = str(target_mask.savepath.savepath) + '.png'

                self._visualize(
                    target_img=target_img,
                    final_mask=target_mask,
                    previous_mask=mask_previous,
                    save_path=save_path,
                    show=visualize
                )
        else:
            print("No sources detected to mask.")
        return target_mask

    def mask_circle(self,
                    # Input parameters
                    target_img: Union[ScienceImage, ReferenceImage, CalibrationImage],
                    target_mask: Optional[Mask] = None,
                    mask_type: str = 'source',
                    x_position: float = None,
                    y_position: float = None,
                    radius_arcsec: float = None,
                    unit: str = 'deg',
                    
                    # Others
                    save: bool = False,
                    verbose: bool = True,
                    visualize: bool = True,
                    save_fig: bool = False,
                    ):
        """
        Add a circular mask to the mask image using photutils CircularAperture.
        
        Parameters
        ----------
        target_img : ScienceImage or ReferenceImage or CalibrationImage
            The target image to generate the circular mask from.
        target_mask : Mask, optional
            The mask to use for the circular mask generation.
        mask_type : str, optional
            The type of mask to generate. ['invalid', 'source', 'cosmicray', 'badpixel', 'subtraction']
        x_position : float, optional
            The x position of the circular mask.
        y_position : float, optional
            The y position of the circular mask.
        radius_arcsec : float, optional
            The radius of the circular mask in arcseconds.
        unit : str, optional
            The unit of the radius. ['deg', 'pixel']
        save : bool, optional
            Whether to save the circular mask.
        verbose : bool, optional
            Whether to print verbose output.
        visualize : bool, optional
            Whether to visualize the circular mask.
        save_fig : bool, optional
            Whether to save the circular mask as a figure.
            
        Returns
        -------
        target_mask : Mask
            The circular mask.
        """
        if target_mask is None:
            if mask_type == 'invalid':
                target_mask = Mask(target_img.savepath.invalidmaskpath, masktype = mask_type, load=False)
            elif mask_type == 'source':
                target_mask = Mask(target_img.savepath.srcmaskpath, masktype = mask_type, load=False)
            elif mask_type == 'cosmicray':
                target_mask = Mask(target_img.savepath.crmaskpath, masktype = mask_type, load=False)
            elif mask_type == 'badpixel':
                target_mask = Mask(target_img.savepath.bpmaskpath, masktype = mask_type, load=False)
            elif mask_type == 'subtraction':
                target_mask = Mask(target_img.savepath.submaskpath, masktype = mask_type, load=False)
            else:
                self.helper.print(f"Unknown mask type: {mask_type}. Using 'invalid' as default.", verbose)
        else:
            self.helper.print("External mask is loaded.", verbose)

        if unit == 'deg':
            if target_img.header is None:
                raise ValueError("Header is required for RA/Dec conversion.")
            w = WCS(target_img.header)
            x_position_pixel, y_position_pixel = w.wcs_world2pix(x_position, y_position, 0)
            pixel_scales_deg = proj_plane_pixel_scales(w)  # [dy, dx] in deg/pixel
            pixel_scale_arcsec = np.mean(pixel_scales_deg) * 3600.0  # arcsec/pixel
            radius_pixel = radius_arcsec / pixel_scale_arcsec  # arcsec ? pixel
        else:
            x_position_pixel, y_position_pixel = x_position, y_position
            radius_pixel = radius_arcsec

        shape = target_img.data.shape
        aperture = CircularAperture((x_position_pixel, y_position_pixel), r=radius_pixel)
        new_mask = aperture.to_mask(method='center').to_image(shape)      

        mask_previous = target_mask.data
        target_mask.combine_mask(new_mask, 'or')
        target_mask.header = target_img.header  
        
        # Update header/status
        update_header_kwargs = dict(
            TGTPATH = str(target_img.path),
            )
        
        current_masktype = target_mask.info.MASKTYPE
        if 'MASKTYPE' not in target_mask.header:
            update_header_kwargs['MASKTYPE'] = "Aperture"
        else:
            if "Aperture" not in current_masktype:
                update_header_kwargs['MASKTYPE'] = f"{current_masktype},Aperture"
        
        ## Update attempt
        if 'MASKATMP' not in target_mask.info.to_dict():
            update_header_kwargs['MASKATMP'] = 1
        else:
            update_header_kwargs['MASKATMP'] = int(target_mask.info.MASKATMP) + 1
        target_mask.header.update(update_header_kwargs)
                    
        ## Update status
        event_details = dict(x=x_position, y=y_position, radius_arcsec=radius_arcsec, unit=unit)
        target_mask.add_status("circular_mask", **event_details)
        self.helper.print(f"Added circular mask at ({x_position:.2f}, {y_position:.2f}) with radius {radius_arcsec}arcsec", verbose)
        
        # Save mask
        if save:
            target_mask.write()
        
        # Visualize
        if visualize or save_fig:
            save_path = None
            if save_fig:
                save_path = str(target_mask.savepath.savepath) + '.png'

            self._visualize(
                target_img=target_img,
                final_mask=target_mask,
                previous_mask=mask_previous,
                save_path=save_path,
                show=visualize
            )
        return target_mask
    
    def mask_cosmicray(self,
                       # Input parameters
                       target_img: Union[ScienceImage, ReferenceImage, CalibrationImage],
                       target_mask: Optional[Mask] = None,
                       gain: float = None,
                       readnoise: float = None,
                       sigclip: float = 6,
                       sigfrac: float = 0.5,
                       objlim: float = 5.0,
                       niter: int = 4,
                       cleantype: str = 'medmask',
                       fsmode: str = 'median',
                       psffwhm: float = None,
                       saturation_level: float = 30000,
                       
                       # Others
                       save: bool = False,
                       verbose: bool = True,
                       visualize: bool = True,
                       save_fig: bool = False,
                       ):
        """
        Generate cosmic ray mask.
        
        Parameters
        ----------
        target_img : ScienceImage or ReferenceImage or CalibrationImage
            The target image to generate the cosmic ray mask from.
        target_mask : Mask, optional
            The mask to use for the cosmic ray mask generation.
        gain : float, optional
            The gain of the image.
        readnoise : float, optional
            The readnoise of the image.
        sigclip : float, optional
            The sigma clip for the cosmic ray mask generation.
        sigfrac : float, optional
            The sigma fraction for the cosmic ray mask generation.
        objlim : float, optional
            The object limit for the cosmic ray mask generation.
        niter : int, optional
            The number of iterations for the cosmic ray mask generation.
        cleantype : str, optional
            The clean type for the cosmic ray mask generation.
        fsmode : str, optional
            The fsmode for the cosmic ray mask generation.
        psffwhm : float, optional
            The psf fwhm for the cosmic ray mask generation.
        saturation_level : float, optional
            The saturation level for the cosmic ray mask generation.
        save : bool, optional
            Whether to save the cosmic ray mask.
        verbose : bool, optional
            Whether to print verbose output.
        visualize : bool, optional
            Whether to visualize the cosmic ray mask.
        save_fig : bool, optional
            Whether to save the cosmic ray mask as a figure.
            
        Returns
        -------
        target_mask : Mask
            The cosmic ray mask.
        """

        # Perform cosmic ray detection and cleaning
        if target_mask is None:
            target_mask = Mask(target_img.savepath.crmaskpath, masktype = 'cosmicray', load=False)
        else:
            self.helper.print("External mask is loaded.", verbose)
        # Load information from target_img
        if gain is None:
            gain = target_img.egain
        if readnoise is None:
            readnoise = target_img.telinfo['readnoise']
        if (gain is None) or (readnoise is None):
            raise ValueError("Gain and readnoise are required for cosmic ray detection.")
        if psffwhm is None:
            psffwhm = 2 / target_img.telinfo['pixelscale']
        
        self.print(f'Detecting cosmic ray... [sigma = {sigclip}, n_iter = {niter}, mode = {fsmode}]', verbose)
        new_mask, clean_image = cr.detect_cosmics(
            target_img.data, gain=gain, readnoise=readnoise, 
            sigclip=sigclip, sigfrac=sigfrac, 
            objlim=objlim, niter=niter, 
            cleantype=cleantype, fsmode=fsmode, 
            psffwhm = psffwhm, verbose=verbose,
            satlevel = saturation_level)
        
        mask_previous = target_mask.data
        target_mask.combine_mask(new_mask, 'or')
        target_mask.header = target_img.header        
        
        # Update header/status
        update_header_kwargs = dict(
            TGTPATH = str(target_img.path),
            )
        
        current_masktype = target_mask.info.MASKTYPE
        if 'MASKTYPE' not in target_mask.header:
            update_header_kwargs['MASKTYPE'] = "CosmicRay"
        else:
            if "CosmicRay" not in current_masktype:
                update_header_kwargs['MASKTYPE'] = f"{current_masktype},CosmicRay"
        
        ## Update attempt
        if 'MASKATMP' not in target_mask.info.to_dict():
            update_header_kwargs['MASKATMP'] = 1
        else:
            update_header_kwargs['MASKATMP'] = int(target_mask.info.MASKATMP) + 1
        target_mask.header.update(update_header_kwargs)
        
        ## Update status
        event_details = dict(gain = gain, readnoise = readnoise, sigclip = sigclip, sigfrac = sigfrac, objlim = objlim, niter = niter, cleantype = cleantype, fsmode = fsmode)
        target_mask.add_status("cr_mask", **event_details)
        self.helper.print(f"{new_mask.sum()} cosmic rays masked.", verbose)
        
        # Save mask
        if save:
            target_mask.write()
        
        # Visualize
        if visualize or save_fig:
            save_path = None
            if save_fig:
                save_path = str(target_mask.savepath.savepath) + '.png'

            self._visualize(
                target_img=target_img,
                final_mask=target_mask,
                previous_mask=mask_previous,
                save_path=save_path,
                show=visualize
            )
        return target_mask
        
    def _visualize(self,
                   target_img: Union[ScienceImage, ReferenceImage, CalibrationImage],
                   final_mask: Optional[Mask],
                   previous_mask: np.ndarray = None,
                   save_path: str = None,
                   show: bool = False):
        """
        Visualize the image and mask.
        """
        from astropy.visualization import ZScaleInterval
        
        interval = ZScaleInterval()
        
        def downsample(data, factor=4):
            return data[::factor, ::factor]
        
        image_data = target_img.data
        image_data_small = downsample(image_data)
        bkg_value = np.mean(image_data_small)
        bkg_rms = np.std(image_data_small)
        if previous_mask is not None:
            previous_mask_small = downsample(previous_mask)
        new_mask = final_mask.data
        new_mask_small = downsample(new_mask)
        len_figure = 1 + sum(mask is not None for mask in [previous_mask, new_mask])
        # Visualization of the image
        
        fig, ax = plt.subplots(1, len_figure, figsize=(6 * len_figure, 6))
        divider = make_axes_locatable(ax[0])
        cax = divider.append_axes('right', size='5%', pad=0.05)
        vmin, vmax = interval.get_limits(image_data_small)
        im0 = ax[0].imshow(image_data_small, origin='lower', cmap='Greys_r', vmin=vmin, vmax=vmax)
        ax[0].set_title('Original Image')
        fig.colorbar(im0, cax=cax, orientation='vertical')
        
        if previous_mask is not None:
            divider = make_axes_locatable(ax[1])
            cax = divider.append_axes('right', size='5%', pad=0.05)
            im1 = ax[1].imshow(previous_mask_small, origin='lower', cmap='Greys_r', vmin=0, vmax=1)
            ax[1].set_title('Previous Mask')
            fig.colorbar(im1, cax=cax, orientation='vertical')
        
        divider = make_axes_locatable(ax[-1])
        cax = divider.append_axes('right', size='5%', pad=0.05)
        im2 = ax[-1].imshow(new_mask_small, origin='lower', cmap='Greys_r', vmin=0, vmax=1)
        ax[-1].set_title('New Mask')
        fig.colorbar(im2, cax=cax, orientation='vertical')
        plt.tight_layout()
        
        if save_path is not None:
            plt.savefig(save_path, dpi=300, bbox_inches='tight')
            
        if show:
            plt.show()
            
        plt.close(fig)




#%%
from typing import List, Union
import glob
from pathlib import Path
from tqdm import tqdm
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import cKDTree
from astropy.table import Table
from astropy.time import Time
from astropy.io import fits
from astropy.coordinates import SkyCoord
import astropy.units as u
from matplotlib import cycler
from itertools import cycle
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize
import matplotlib.cm as cm



from ezphot.dataobjects import Catalog, CatalogSet
from ezphot.helper import Helper
from ezphot.imageobjects import ScienceImage
from ezphot.utils import SDTDataQuerier
from ezphot.utils import CatalogQuerier

#%%
class PhotometricSpectrum:
    """
    Photometric spectrum object.
    
    This object is used to plot the photometric spectrum of a source.
    It is initialized with a source_catalogs object, which is a CatalogSet object.
    The source_catalogs object is used to extract the source information from the catalogs.
    
    To change the plot parameters, modify the plt_params attribute.
    """
    
    def __init__(self, source_catalogs: CatalogSet = None):
        """
        Initialize the PhotometricSpectrum object.
        
        Parameters
        ----------
        source_catalogs : CatalogSet
            The source catalogs to use for the photometric spectrum.
        """
        if not isinstance(source_catalogs, CatalogSet):
            raise TypeError("source_catalogs must be an instance of CatalogSet")
        self.helper = Helper()
        self.source_catalogs = source_catalogs
        self.catalogs_tbl = None
        self.plt_params = self._plt_params()
        self.CatalogQuerier = CatalogQuerier(catalog_key = None)
        self.data = None
        
    OFFSET = 2
        
    EFFECTIVE_WAVELENGTHS_NM = {
        'm400': 400.0,
        'm412': 412.5,
        'm425': 425.0,
        'm437': 437.5,
        'm450': 450.0,
        'm462': 462.5,
        'm475': 475.0,
        'm487': 487.5,
        'm500': 500.0,
        'm512': 512.5,
        'm525': 525.0,
        'm537': 537.5,
        'm550': 550.0,
        'm562': 562.5,
        'm575': 575.0,
        'm587': 587.5,
        'm600': 600.0,
        'm612': 612.5,
        'm625': 625.0,
        'm637': 637.5,
        'm650': 650.0,
        'm662': 662.5,
        'm675': 675.0,
        'm687': 687.5,
        'm700': 700.0,
        'm712': 712.5,
        'm725': 725.0,
        'm737': 737.5,
        'm750': 750.0,
        'm762': 762.5,
        'm775': 775.0,
        'm787': 787.5,
        'm800': 800.0,
        'm812': 812.5,
        'm825': 825.0,
        'm837': 837.5,
        'm850': 850.0,
        'm862': 862.5,
        'm875': 875.0,
        'm887': 887.5,
        # SDSS ugriz (https://mfouesneau.github.io/pyphot/libcontent.html)
        'u': 355.7,
        'g': 470.2,
        'r': 617.6,
        'i': 749.0,
        'z': 894.7,
        # PS1 ugizy (https://mfouesneau.github.io/pyphot/libcontent.html)
        'g_ps1': 484.9,
        'r_ps1': 620.2,
        'i_ps1': 753.5,
        'z_ps1': 867.4,
        'y_ps1': 962.8,
        # Johnson-Cousins UBVRI (Ground based, https://mfouesneau.github.io/pyphot/libcontent.html)
        'U': 363.5,
        'B': 429.7,
        'V': 547.0,
        'R': 647.1,
        'I': 787.2,
        # 2MASS JHK (Ground based, https://mfouesneau.github.io/pyphot/libcontent.html)
        'J': 1230.3,
        'H': 1640.3,
        'K': 2202.7,
        # WISE W1-W4 (https://mfouesneau.github.io/pyphot/libcontent.html)
        'W1': 3368.0,
        'W2': 4618.0,
        'W3': 12073.0,
        'W4': 22194.0,
    }
    
    FILTER_SHAPE = {
        'm400': 'o',
        'm412': 'o',
        'm425': 'o',
        'm437': 'o',
        'm450': 'o',
        'm462': 'o',
        'm475': 'o',
        'm487': 'o',
        'm500': 'o',
        'm512': 'o',
        'm525': 'o',
        'm537': 'o',
        'm550': 'o',
        'm562': 'o',
        'm575': 'o',
        'm587': 'o',
        'm600': 'o',
        'm612': 'o',
        'm625': 'o',
        'm637': 'o',
        'm650': 'o',
        'm662': 'o',
        'm675': 'o',
        'm687': 'o',
        'm700': 'o',
        'm712': 'o',
        'm725': 'o',
        'm737': 'o',
        'm750': 'o',
        'm762': 'o',
        'm775': 'o',
        'm787': 'o',
        'm800': 'o',
        'm812': 'o',
        'm825': 'o',
        'm837': 'o',
        'm850': 'o',
        'm862': 'o',
        'm875': 'o',
        'm887': 'o',
        # SDSS ugriz
        'u': 's',
        'g': 's',
        'r': 's',
        'i': 's',
        'z': 's',
        # PS1 ugizy
        'g_ps1': 's',
        'r_ps1': 's',
        'i_ps1': 's',
        'z_ps1': 's',
        'y_ps1': 's',
        # Johnson-Cousins UBVRI
        'U': 's',
        'B': 's',
        'V': 's',
        'R': 's',
        'I': 's',
        # 2MASS JHK
        'J': 's',
        'H': 's',
        'K': 's',
        # WISE W1-W4
        'W1': 's',
        'W2': 's',
        'W3': 's',
        'W4': 's',
    }
    
    FILTER_COLOR = {
        'u': 'cyan',
        'g': 'green',
        'r': 'red',
        'i': 'black',         
        'z': 'brown',         
        'y': 'darkorange',    
        'B': 'royalblue',
        'V': 'limegreen',
        'R': 'firebrick',
        'I': 'maroon',
    } 
    
    def __repr__(self):
        txt = f'PHOTOMETRIC SPECTRUM OBJECT (n_catalogs = {len(self.source_catalogs.catalogs)})\n'
        txt += str(self.plt_params)
        return txt
    
    def plot(self, 
             ra: float,
             dec: float,
             objname: str = None,
             matching_radius_arcsec: float = 5.0,
             flux_key: str = 'MAGSKY_APER_1',
             fluxerr_key: str = 'MAGERR_APER_1',
             color_key: str = 'OBSDATE',        # 'OBSDATE' or 'OBSERVATORY'
             overplot_gaiaxp: bool = False,
             overplot_sdss: bool = False,
             overplot_ps1: bool = False,
             verbose: bool = True,
            ):
        """
        Plot photometric spectrum (wavelength vs magnitude/flux) for the given source.
        
        The figure parameters are set in the plt_params attribute.
        
        Parameters
        ----------
        ra : float
            Right ascension of the source in degrees.
        dec : float
            Declination of the source in degrees.
        objname : str, optional
            Name of the source.
        matching_radius_arcsec : float, optional
            Matching radius in arcseconds.
        flux_key : str, optional
            Key for the flux column.
        fluxerr_key : str, optional
            Key for the flux error column.
        color_key : str, optional
            Key for the color column.
        overplot_gaiaxp : bool, optional
            Whether to overplot GaiaXP data.
        overplot_sdss : bool, optional
            Whether to overplot SDSS data.
        overplot_ps1 : bool, optional
            Whether to overplot PS1 data.
        verbose : bool, optional
            Whether to print verbose output.
            
        Returns
        -------
        figs : list
            List of figures.
        axs : list
            List of axes.
        tbl : astropy.table.Table
            Table of data.
        """
        # Ensure per-exposure table exists
        if self.data is None:
            self.extract_source_info(
                ra, dec,
                flux_keys=[flux_key],
                fluxerr_keys=[fluxerr_key],
                matching_radius_arcsec=matching_radius_arcsec
            )
        if self.data is None or len(self.data) == 0:
            self.helper.print(f"[WARNING] No sources found within {matching_radius_arcsec}\" of RA={ra}, Dec={dec}", verbose)
            return None, None, None
        
        tbl = self.data.copy()

        # Basic info & title
        coord = tbl['coord'][0] if 'coord' in tbl.colnames else SkyCoord(ra*u.deg, dec*u.deg)
        ra_str  = coord.ra.to_string(unit=u.hourangle, sep='', pad=True, precision=2)
        dec_str = coord.dec.to_string(sep='', alwayssign=True, pad=True, precision=1)
        jname   = f'J{ra_str}{dec_str}'
        title   = objname if objname is not None else jname

        # labels / grouping
        if color_key.lower() == 'obsdate':
            # 1) ensure a 'group' column exists (whatever your group_table does)
            tbl = self.helper.group_table(tbl, 'mjd')        # must add/keep tbl['group']
            gview = tbl.group_by('group')

            # 2) compute a mean-MJD label per group
            key_vals = np.array(gview.groups.keys['group'])  # one key per group
            labels_map = {}
            for i, g in enumerate(gview.groups):
                mjd_mean = float(np.nanmean(g['mjd']))
                date_str = Time(mjd_mean, format='mjd').to_value('iso', subfmt='date_hm')
                labels_map[key_vals[i]] = date_str

            # 3) propagate to all rows
            obsdate_group = np.empty(len(tbl), dtype=object)
            for k, lab in labels_map.items():
                obsdate_group[tbl['group'] == k] = lab
            tbl['obsdate_group'] = obsdate_group

            # 4) this is the per-row groups array you sort alongside wl/vals/errs
            groups = np.array(tbl['obsdate_group'], dtype=object)

        elif color_key.lower() == 'observatory':
            groups = np.array(tbl['observatory'], dtype=object)
            color_key = 'OBSERVATORY'
        
        elif color_key.lower() == 'telname':
            groups = np.array(tbl['telname'], dtype=object)
            color_key = 'TELNAME'
        
        else:
            self.helper.print(f"[WARNING] Invalid color_key: {color_key}", verbose)
            return None, None, None
            
        # Build arrays per row
        # wavelength
        wl = np.array([self._band_to_wavelength_nm(b) for b in tbl['filter']], dtype=float)
        vals  = np.array(tbl[flux_key], dtype=float)
        merrs = np.array(tbl[fluxerr_key], dtype=float)
        zperrs = np.array(tbl['zp_err'], dtype=float)
        errs  = np.array([self._combine_err(m, z) for m, z in zip(merrs, zperrs)], dtype=float)

        # Sort by wavelength for nicer lines
        order = np.argsort(wl)
        wl, vals, errs, groups = wl[order], vals[order], errs[order], np.array(groups)[order]

        # BEFORE sorting:
        wl   = np.array([self._band_to_wavelength_nm(b) for b in tbl['filter']], dtype=float)
        vals = np.array(tbl[flux_key], dtype=float)
        merrs = np.array(tbl[fluxerr_key], dtype=float)
        zperrs = np.array(tbl['zp_err'], dtype=float)
        errs = np.array([self._combine_err(m, z) for m, z in zip(merrs, zperrs)], dtype=float)
        filt = np.array(tbl['filter'], dtype=object)  # <-- keep filters as an array

        # Sort everything together:
        order = np.argsort(wl)
        wl, vals, errs, groups, filt = wl[order], vals[order], errs[order], np.array(groups)[order], filt[order]


        unique_groups = list(dict.fromkeys(groups))  # preserve order
        self.plt_params.set_palette(n=len(unique_groups))
        with self.plt_params.apply():
            # Group colors: use observation time as a continuous colorbar when grouping by date
            
            if color_key.lower() == 'obsdate':
                group_mjds = [Time(g).mjd for g in unique_groups]
                norm = Normalize(vmin=np.nanmin(group_mjds), vmax=np.nanmax(group_mjds))
                cmap_name = getattr(self.plt_params, 'cmap', None)
                if cmap_name is None:
                    cmap_name = 'jet'
                cmap = cm.get_cmap(cmap_name)
                colors = [cmap(norm(x)) for x in group_mjds]
            else:
                # categorical palette
                base_cycle = cycle(plt.rcParams['axes.prop_cycle'].by_key()['color'])
                colors = [next(base_cycle) for _ in unique_groups]
                
            # dynamic height based on #groups and your offset
            # dynamic height
            fig_height = 3 + len(unique_groups) * 0.35 * self.OFFSET

            # use your chosen width and dpi from plt_params
            width, _ = self.plt_params.figure_figsize
            self.plt_params._rcparams['figure.figsize'] = (width, fig_height)
            self.plt_params._rcparams['figure.dpi'] = self.plt_params.figure_dpi

            fig, ax = plt.subplots()  # no figsize here

            offset = 0.0
            # for cg, col in zip(unique_groups, colors):
            #     m = (groups == cg)
            #     x = wl[m]
            #     y = vals[m].copy()
            #     e = errs[m]

            #     y = y + offset
            #     valid = np.isfinite(x) & np.isfinite(y)
            #     if np.any(valid):
            #         ax.plot(x[valid], y[valid], c=col, alpha=0.5, linestyle=self.plt_params.line_style)
            #         ax.errorbar(x[valid], y[valid], yerr=e[valid],
            #                     label=f"{cg} (+{offset:.1f})" if offset != 0 else f"{cg}",
            #                     **self.plt_params.get_errorbar_kwargs(color=col, shape='o'))
            #     offset += self.OFFSET
            
            has_medium = False  # circle 'o'
            has_broad = False   # square 's'
            for cg, col in zip(unique_groups, colors):
                m = (groups == cg)
                x = wl[m]
                y = vals[m].copy()
                e = errs[m]
                f = filt[m]

                y = y + offset
                valid = np.isfinite(x) & np.isfinite(y)
                if np.any(valid):
                    ax.plot(x[valid], y[valid], c=col, alpha=0.5,
                            linestyle=self.plt_params.line_style)

                    first_point = True
                    for xi, yi, ei, ff in zip(x[valid], y[valid], e[valid], f[valid]):
                        shape = self.FILTER_SHAPE.get(ff, 'o')
                        color = self.FILTER_COLOR.get(ff, col)

                        # Track shape usage
                        if shape == 'o':
                            has_medium = True
                        elif shape == 's':
                            has_broad = True

                        if first_point:
                            ax.errorbar(
                                xi, yi, yerr=ei,
                                label=f"{cg} (+{offset:.1f})" if offset != 0 else f"{cg}",
                                **self.plt_params.get_errorbar_kwargs(color=color, shape=shape)
                            )
                            first_point = False
                        else:
                            ax.errorbar(
                                xi, yi, yerr=ei,
                                **self.plt_params.get_errorbar_kwargs(color=color, shape=shape)
                            )

                offset += self.OFFSET
            
            # --- Conditionally add black legend entries ---
            if has_medium:
                ax.errorbar([], [], **self.plt_params.get_errorbar_kwargs(color='k', shape='o'), label='Medium bands')
            if has_broad:
                ax.errorbar([], [], **self.plt_params.get_errorbar_kwargs(color='k', shape='s'), label='Broad bands')

            # Axes / style
            ax.set_xlabel("Effective Wavelength [nm]")
            ax.set_ylabel("Magnitude (+ offset)" if "MAG" in flux_key.upper() else "Flux (+ offset)")
            if "MAG" in flux_key.upper():
                ax.invert_yaxis()
            if self.plt_params.xlim:  ax.set_xlim(*self.plt_params.xlim)
            if self.plt_params.ylim:  ax.set_ylim(*self.plt_params.ylim)
            if self.plt_params.xticks is not None: ax.set_xticks(self.plt_params.xticks)
            if self.plt_params.yticks is not None: ax.set_yticks(self.plt_params.yticks)
            ax.grid(True, which='major', alpha=0.3)
            ax.minorticks_on()
            ax.set_title(f"Photometric spectrum for {title}")
            
            # ---------- External overplots ----------
            is_mag = ("MAG" in flux_key.upper())

            # GaiaXP: full low-res spectrum converted to AB mag vs nm
            if overplot_gaiaxp and is_mag:
                try:
                    self.CatalogQuerier.change_catalog('GAIAXP')
                    res = self.CatalogQuerier.query(coord=coord, radius_arcsec=matching_radius_arcsec)
                    if len(res) > 0:
                        gx = res[0]
                        # nearest source
                        closest = gx['Source'][0]
                        gx = gx[gx['Source'] == closest]
                        wl_nm = np.array(gx['lambda'])
                        f_si  = np.array(gx['Flux'])
                        fe_si = np.array(gx['e_Flux'])
                        wl_AA = wl_nm * 10.0
                        mag   = np.array(self.helper.flambSI_to_ABmag(f_si, wl_AA), dtype=float)
                        magerr = np.array(self.helper.fluxerr_to_magerr(flux=f_si, fluxerr=fe_si), dtype=float)
                        ok = np.isfinite(mag) & (magerr >= 0)
                        if np.any(ok):
                            ax.errorbar(wl_nm[ok], mag[ok], yerr=magerr[ok],
                                        fmt='None', color='magenta', alpha=0.3, label='GaiaXP')
                    else:
                        ax.plot([], [], ' ', label='GaiaXP (no data)')
                except Exception:
                    ax.plot([], [], ' ', label='GaiaXP (error)')

            # SDSS points (u,g,r,i,z)
            if overplot_sdss and is_mag:
                try:
                    self.CatalogQuerier.change_catalog('SDSS')
                    res = self.CatalogQuerier.query(coord=coord, radius_arcsec=matching_radius_arcsec)
                    if len(res) > 0 and len(res[0]) > 0:
                        sdss = res[0][0]
                        bands = ['u','g','r','i','z']
                        xs, ys, es = [], [], []
                        for b in bands:
                            m  = sdss.get(f'{b}mag')
                            me = sdss.get(f'e_{b}mag')
                            wl = self._band_to_wavelength_nm(b)
                            if m is None or me is None or not np.isfinite(wl):
                                continue
                            xs.append(wl); ys.append(m); es.append(me)
                        if xs:
                            ax.errorbar(xs, ys, yerr=es,
                                        label='SDSS', **self.plt_params.get_errorbar_kwargs('green','^'))
                    else:
                        ax.plot([], [], ' ', label='SDSS (no data)')
                except Exception:
                    ax.plot([], [], ' ', label='SDSS (error)')

            # PS1 points (g,r,i,z,y)
            if overplot_ps1 and is_mag:
                try:
                    self.CatalogQuerier.change_catalog('PS1')
                    res = self.CatalogQuerier.query(coord=coord, radius_arcsec=matching_radius_arcsec)
                    if len(res) > 0 and len(res[0]) > 0:
                        ps1 = res[0][0]
                        bands = ['g','r','i','z','y']
                        xs, ys, es = [], [], []
                        for b in bands:
                            m  = ps1.get(f'{b}mag')
                            me = ps1.get(f'e_{b}mag')
                            wl = self._band_to_wavelength_nm(f'{b}_ps1')  # use PS1-specific pivot
                            if (m is None) or (me is None) or (not np.isfinite(wl)):
                                continue
                            xs.append(wl); ys.append(m); es.append(me)
                        if xs:
                            ax.errorbar(xs, ys, yerr=es,
                                        label='PS1', **self.plt_params.get_errorbar_kwargs('blue','o'))
                    else:
                        ax.plot([], [], ' ', label='PS1 (no data)')
                except Exception:
                    ax.plot([], [], ' ', label='PS1 (error)')

            if (len(unique_groups) > 1) or any([overplot_gaiaxp, overplot_sdss, overplot_ps1]):
                ax.legend(loc=self.plt_params.label_position, ncol=self.plt_params.ncols,
                        fontsize=self.plt_params.legend_fontsize)

            plt.show()
            return [fig], [ax], tbl

    def search_source(self, 
                      ra: Union[float, list, np.ndarray],
                      dec: Union[float, list, np.ndarray],
                      ra_key: str = 'X_WORLD',
                      dec_key: str = 'Y_WORLD',
                      matching_radius_arcsec: float = 5.0):
        """
        Search for sources in the merged catalog.
        
        Parameters
        ----------
        ra, dec : float
            Sky position in degrees.
        ra_key : str
            Column name for right ascension. Default is 'X_WORLD'.
        dec_key : str
            Column name for declination. Default is 'Y_WORLD'.
        matching_radius_arcsec : float
            Search radius for the source match.

        """
        if self.merged_catalog is None:
            self._merge_catalogs(ra_key = ra_key, 
                                 dec_key = dec_key,
                                 max_distance_arcsec=matching_radius_arcsec,
                                 join_type = 'outer',
                                 )

        ra = np.atleast_1d(ra)
        dec = np.atleast_1d(dec)
        input_coords = SkyCoord(ra=ra, dec=dec, unit='deg')
        
        target_catalog = self.merged_catalog
        catalog_coords = target_catalog['coord']
        
        matched_catalog, matched_input, unmatched_catalog = self.helper.cross_match(catalog_coords, input_coords, matching_radius_arcsec)
        print(f"Matched {len(matched_catalog)} sources out of {len(input_coords)} input positions.")
        return target_catalog[matched_catalog]

    def extract_source_info(self,
                            ra: float,
                            dec: float,
                            ra_key: str = 'X_WORLD',
                            dec_key: str = 'Y_WORLD',
                            flux_keys=['MAGSKY_AUTO', 'MAGSKY_APER', 'MAGSKY_APER_1', 'MAGSKY_APER_2', 'MAGSKY_APER_3', 'MAGSKY_APER_4'], 
                            fluxerr_keys=['MAGERR_AUTO', 'MAGERR_APER', 'MAGERR_APER_1', 'MAGERR_APER_2', 'MAGERR_APER_3', 'MAGERR_APER_4'], 
                            matching_radius_arcsec=5.0, 
                            fit_filter_key=None):
        """
        Extract source information from the merged catalog.
        
        Each row of the returned table will be a per-exposure record with the metadata and photometry.
        
        Parameters
        ----------
        ra, dec : float
            Sky position in degrees.
        ra_key : str
            Column name for right ascension. Default is 'X_WORLD'.
        dec_key : str
            Column name for declination. Default is 'Y_WORLD'.
        flux_keys : str or sequence of str
            Photometry value columns you want to carry over (e.g., 'MAGSKY_APER_1').
            Accepts a single key or a list/tuple of keys.
        fluxerr_keys : str or sequence of str
            Corresponding error columns (e.g., 'MAGERR_APER_1'). Must be same length as `flux_keys`.
        matching_radius_arcsec : float
            Search radius for the source match.
        fit_filter_key : str or None
            Placeholder for future use (ignored).

        Returns
        -------
        astropy.table.Table or None
            One row per exposure/catalog with metadata + requested photometry.
            Returns None if no source is found.
        """
        # Normalize keys to lists
        if isinstance(flux_keys, str):
            flux_keys = [flux_keys]
        if isinstance(fluxerr_keys, str):
            fluxerr_keys = [fluxerr_keys]
        if len(flux_keys) != len(fluxerr_keys):
            raise ValueError("flux_keys and fluxerr_keys must have the same length.")

        # Ensure merged table exists (pulling at least the requested keys + their ZPERR counterparts)
        needed_data_keys = set()
        for fk, fek in zip(flux_keys, fluxerr_keys):
            needed_data_keys.add(fk)
            needed_data_keys.add(fek)
            needed_data_keys.add(fek.replace('MAGERR', 'ZPERR'))

        if getattr(self, "catalogs_tbl", None) is None:
            # Provide sane defaults for merge (user can override upstream if needed)
            self._merge_catalogs(
                ra_key = ra_key,
                dec_key = dec_key,
                max_distance_arcsec=matching_radius_arcsec,
                join_type = 'outer',
                data_keys=sorted(needed_data_keys),
            )
        else:
            # If any needed key is missing, re-merge including them
            need_merge = False
            for key in needed_data_keys:
                if f"{key}_idx0" not in self.merged_catalog.colnames:
                    need_merge = True
                    break
            if need_merge:
                self._merge_catalogs(
                    ra_key = ra_key,
                    dec_key = dec_key,
                    max_distance_arcsec=matching_radius_arcsec,
                    join_type = 'outer',
                    data_keys=sorted(needed_data_keys),
                )

        # Find closest source
        selected = self.search_source(ra, 
                                      dec, 
                                      ra_key = ra_key,
                                      dec_key = dec_key,
                                      matching_radius_arcsec=matching_radius_arcsec)
        if selected is None or len(selected) == 0:
            return None

        # Take the nearest match (row 0)
        row = selected[0]

        # Build per-exposure records from self.metadata indices
        # self.metadata[idx] should include per-catalog fields like filter, obsdate, depth, observatory, telname, exptime, etc.
        if not hasattr(self, "metadata") or self.metadata is None:
            raise RuntimeError("self.metadata is missing. Run merge_catalogs() first.")

        records = {
            idx: {k: v for k, v in meta.items() if k not in ('ra', 'dec')}
            for idx, meta in self.metadata.items()
        }

        # Copy all "*_idx{idx}" values from the matched row into each record
        # (This will include MAG*, MAGERR*, ZPERR*, and any other requested per-exposure columns.)
        for colname in row.colnames:
            if '_idx' not in colname:
                continue
            try:
                base, idx_str = colname.rsplit('_idx', 1)
                idx = int(idx_str)
            except Exception:
                continue
            if idx in records:
                records[idx][base] = row[colname]

        # For convenience, also add a unified 'zp_err' per exposure using the first available ZPERR among requested pairs
        for idx, rec in records.items():
            zperr_val = None
            for ferr in fluxerr_keys:
                zp_key = ferr.replace('MAGERR', 'ZPERR')
                if zp_key in rec:
                    zperr_val = rec[zp_key]
                    break
            rec['zp_err'] = zperr_val

        # Materialize table
        result_tbl = Table(rows=list(records.values()))

        # Attach coord/ra/dec convenience
        if 'coord' in row.colnames:
            result_tbl['coord'] = row['coord']
        else:
            sc = SkyCoord(ra=ra*u.deg, dec=dec*u.deg)
            result_tbl['coord'] = [sc] * len(result_tbl)

        if 'ra' not in result_tbl.colnames:
            result_tbl['ra'] = float(ra)
        if 'dec' not in result_tbl.colnames:
            result_tbl['dec'] = float(dec)

        # Time columns
        if 'obsdate' in result_tbl.colnames:
            t = Time(result_tbl['obsdate'])
            result_tbl['mjd'] = t.mjd
            result_tbl['jd'] = t.jd

        # Column ordering: metadata ? requested photometry ? remaining
        meta_order = ['ra', 'dec', 'coord', 'filter', 'exptime', 'obsdate', 'mjd', 'jd',
                    'seeing', 'depth', 'observatory', 'telname', 'zp_err']
        phot_cols = []
        for fk, fek in zip(flux_keys, fluxerr_keys):
            if fk in result_tbl.colnames:
                phot_cols.append(fk)
            if fek in result_tbl.colnames:
                phot_cols.append(fek)
            zpk = fek.replace('MAGERR', 'ZPERR')
            if zpk in result_tbl.colnames:
                phot_cols.append(zpk)

        ordered = [c for c in meta_order if c in result_tbl.colnames] + phot_cols
        remaining = [c for c in result_tbl.colnames if c not in ordered]
        result_tbl = result_tbl[ordered + remaining]

        # Cache for plotting
        self.data = result_tbl
        return result_tbl


    def _band_to_wavelength_nm(self, band: str) -> float:
        """Return effective wavelength (nm) for a band key."""
        # exact key
        if band in self.EFFECTIVE_WAVELENGTHS_NM:
            return self.EFFECTIVE_WAVELENGTHS_NM[band]
        # PS1 commonly comes as 'g','r','i','z','y' from services; map to *_ps1
        ps1_map = {'g':'g_ps1','r':'r_ps1','i':'i_ps1','z':'z_ps1','y':'y_ps1'}
        if band in ps1_map and ps1_map[band] in self.EFFECTIVE_WAVELENGTHS_NM:
            return self.EFFECTIVE_WAVELENGTHS_NM[ps1_map[band]]
        return np.nan

    def _combine_err(self, meas_err, zp_err):
        """Quadrature-combine measurement and zeropoint error when both finite."""
        m = meas_err if np.isfinite(meas_err) else np.nan
        z = zp_err   if np.isfinite(zp_err)   else np.nan
        if np.isfinite(m) and np.isfinite(z):
            return np.sqrt(m*m + z*z)
        return m

    def _merge_catalogs(self,
                        ra_key: str = 'X_WORLD',
                        dec_key: str = 'Y_WORLD',
                        max_distance_arcsec: float = 2,
                        join_type: str = 'outer',
                        data_keys: list = ['MAGSKY_AUTO', 'MAGERR_AUTO', 'MAGSKY_APER', 'MASERR_APER', 'MAGSKY_APER_1', 'MAGERR_APER_1', 'MAGSKY_APER_2', 'MAGERR_APER_2', 'MAGSKY_APER_3', 'MAGERR_APER_3', 'MAGSKY_CIRC', 'MAGERR_CIRC']):
        self.merged_catalog, self.metadata = self.source_catalogs.merge_catalogs(
            max_distance_arcsec=max_distance_arcsec,
            ra_key=ra_key,
            dec_key=dec_key,
            join_type=join_type,
            data_keys=data_keys)
    
    def _plt_params(self):
        class PlotParams: 
            def __init__(self):
                self._rcparams = {
                    'figure.figsize': (20, 12),
                    'figure.dpi': 100,
                    'savefig.dpi': 300,
                    'font.family': 'serif',
                    'mathtext.fontset': 'cm',
                    'axes.titlesize': 16,
                    'axes.labelsize': 14,
                    'axes.xmargin': 0.1,
                    'axes.ymargin': 0.2,
                    'axes.prop_cycle': cycler(color=[
                        'black', 'blue', 'orange', 'green', 'red', 'purple', 'brown', 'pink', 'gray',
                        'olive', 'cyan', 'navy', 'gold', 'teal', 'coral', 'darkgreen', 'magenta'
                    ]),
                    'xtick.labelsize': 12,
                    'ytick.labelsize': 12,
                    'legend.fontsize': 9,
                    'lines.linewidth': 1.5,
                    'lines.markersize': 6,
                    'errorbar.capsize': 3,
                    'axes.grid': True,
                    'grid.alpha': 0.3,
                    'xtick.direction': 'in',
                    'ytick.direction': 'in',
                    'xtick.top': True,
                    'ytick.right': True,

                    }
                # Custom axis control
                self.xlim = [350, 925]
                self.ylim = None
                self.xticks = np.arange(400, 901, 50)
                self.yticks = None
                
                # Color parameters
                self.cmap = 'jet'
                
                # Label parameters
                self.label_position = 'best'  # 'best', 'upper right', 'lower left', etc.
                self.ncols = 2
                
                # Error bar parameters
                self.errorbar_enabled = True  # Optional switch
                self.errorbar_markersize = 7
                self.errorbar_hollow_marker = True  # True = hollow, False = filled
                self.errorbar_capsize = 3.5
                self.errorbar_elinewidth = 1.2
                
                # Plot parameters
                self.line_style = 'solid'
                

            def __getattr__(self, name):
                rc_name = name.replace('_', '.')
                if rc_name in self._rcparams:
                    return self._rcparams[rc_name]
                raise AttributeError(f"'PlotParams' object has no attribute '{name}'")

            def __setattr__(self, name, value):
                if name.startswith('_') or name in ('xlim', 'ylim', 'xticks', 'yticks', 'errorbar_capsize', 'errorbar_elinewidth', 'errorbar_markersize', 'errorbar_enabled', 'errorbar_hollow_marker', 'label_position', 'ncols', 'line_style', 'cmap'):
                    super().__setattr__(name, value)
                else:
                    rc_name = name.replace('_', '.')
                    if rc_name in self._rcparams:
                        self._rcparams[rc_name] = value
                    else:
                        raise AttributeError(f"'PlotParams' has no rcParam '{rc_name}'")
                    
            def set_palette(self, n: int = 40):
                base = cm.get_cmap(self.cmap, n)
                colors = [base(i) for i in range(base.N)]
                from matplotlib import cycler as _cycler
                self._rcparams['axes.prop_cycle'] = _cycler(color=colors)
            
            def get_errorbar_kwargs(self, color, shape: str = None):     
                errorbar_kwargs = dict(
                    capsize=self.errorbar_capsize,
                    elinewidth=self.errorbar_elinewidth,
                    markersize=self.errorbar_markersize,
                )
                errorbar_kwargs['mec'] = color
                errorbar_kwargs['color'] = color
                errorbar_kwargs['fmt'] = shape
                
                if self.errorbar_hollow_marker is True:
                    errorbar_kwargs['mfc'] = 'none'
                else:
                    errorbar_kwargs['mfc'] = color

                if self.errorbar_enabled is False:
                    errorbar_kwargs['elinewidth'] = 0
                    errorbar_kwargs['capsize'] = 0
                    
                return errorbar_kwargs
            
            def update(self, **kwargs):
                self._rcparams.update(kwargs)

            def apply(self):
                import matplotlib.pyplot as plt
                return plt.rc_context(self._rcparams)

            def __repr__(self):
                txt = 'PLOT CONFIGURATION ============\n'
                for k, v in self._rcparams.items():
                    txt += f"{k.replace('.', '_')} = {v}\n"
                txt += 'Axis Limits and Ticks -----------\n'
                txt += f"xlim   = {self.xlim}\n"
                txt += f"ylim   = {self.ylim}\n"
                txt += f"xticks = {self.xticks}\n"
                txt += f"yticks = {self.yticks}\n"
                txt += 'Visualization Parameters -----------------\n'
                txt += f"line_style = {self.line_style}\n"
                txt += f"cmap = {self.cmap}\n"
                txt += 'Error Bar Configuration ---------\n'
                txt += f"errorbar_enabled = {self.errorbar_enabled}\n"
                txt += f"errorbar_markersize = {self.errorbar_markersize}\n"
                
                txt += f"errorbar_capsize = {self.errorbar_capsize}\n"
                txt += f"errorbar_elinewidth = {self.errorbar_elinewidth}\n"
                
                return txt
        return PlotParams()

        
        
    
#%%
if __name__ == "__main__":
    from ezphot.dataobjects import CatalogSet
    source_catalogs = CatalogSet()
    source_catalogs.search_catalogs(
        target_name = 'T01358',
        search_key = 'calib*com.fits.cat'
     )    
    ra  =239.03762083
    dec = -68.41128803
    source_catalogs.select_sources(ra, dec, radius = 10)
#%%
if __name__ == "__main__":
    source_catalogs.select_catalogs(obs_start = '2025-02-12', obs_end = '2025-02-13')
    self = PhotometricSpectrum(source_catalogs)
    self.OFFSET = 0
    self.extract_source_info(ra, dec)
#%%
# %%
if __name__ == "__main__":
    self.plt_params.figure_figsize = (10, 6)
    self.plt_params.figure_dpi = 500
    self.plt_params.cmap = 'jet'
    flux_key = 'MAGSKY_APER_2'
    fluxerr_key = 'MAGERR_APER_2'
    matching_radius_arcsec = 10
    color_key: str = 'telname'
    overplot_gaiaxp = True
    overplot_sdss = True
    overplot_ps1 = True
    self.plt_params.line_style = 'none'
    figs, axs, matched_sources = self.plot(ra, 
                          dec, 
                          flux_key=flux_key, 
                          fluxerr_key =fluxerr_key,
                          color_key = color_key, 
                          matching_radius_arcsec=matching_radius_arcsec,
                          overplot_gaiaxp=overplot_gaiaxp,
                          overplot_sdss = overplot_sdss,
                          overplot_ps1 = overplot_ps1,
                          objname = 'SN 2025fvw')
# %%



#%%
from pathlib import Path
import json
from typing import Union

class TIPConfig:
    def __init__(self,
                 telkey: str = None,
                 configpath: Union[Path, str] = Path(__file__).resolve().parent,
                 **kwargs):
        self.telkey = telkey
        self.config = dict()

        # global config params
        self.path_global = Path(configpath)
        self.path_home = Path.home()
        self._configfiles_global = list(self.path_global.glob('*.config'))
        config_global = self._load_configuration(self._configfiles_global)
        self.config.update(config_global)

        if self.telkey:
            self.path_telescope = self.path_global / self.telkey
            self._configfiles_unit = list(self.path_telescope.glob('*.config'))
            if not self._configfiles_unit:
                print('No configuration file is found.\nTo make default configuration files, run tcspy.configuration.make_config')
            else:
                config_unit = self._load_configuration(self._configfiles_unit)
                self.config.update(config_unit)

    def _load_configuration(self, configfiles):
        all_config = dict()
        for configfile in configfiles:
            with open(configfile, 'r') as f:
                config = json.load(f)
                all_config.update(config)
        return all_config

    def _make_configfile(self, dict_params: dict, filename: str, savepath: Union[str, Path]):
        filepath = Path(savepath) / filename
        with open(filepath, 'w') as f:
            json.dump(dict_params, f, indent=4)
        print(f'New configuration file made: {filepath}')

    def _initialize_config(self):
        savepath_tel = self.path_telescope
        savepath_global = self.path_global

        savepath_tel.mkdir(parents=True, exist_ok=True)

        # LOCAL CONFIGURATION
        sex_config = dict(
            SEX_CONFIG = str(self.path_global / 'sextractor' / f'{self.telkey}.sexconfig'),
            SEX_CONFIGDIR = str(self.path_global / 'sextractor'),
            SEX_LOGDIR = str(self.path_home / 'code' / 'sextractor' / 'log'),
            SEX_HISTORYDIR = str(self.path_home / 'code' / 'sextractor' / 'history')
        )
        astrometry_config = dict(
            ASTROMETRY_SEXCONFIG = str(self.path_global / 'sextractor' / f'{self.telkey}.astrometry.sexconfig')
            )
        scamp_config = dict(
            SCAMP_CONFIG = str(self.path_global / 'scamp' / 'default.scampconfig'),
            SCAMP_SEXCONFIG = str(self.path_global / 'sextractor' / f'{self.telkey}.scamp.sexconfig'),
            SCAMP_CONFIGDIR = str(self.path_global / 'scamp'),
            SCAMP_LOGDIR = str(self.path_home / 'code' / 'scamp' / 'log'),
            SCAMP_HISTORYDIR = str(self.path_home / 'code' / 'scamp' / 'history')
        )
        swarp_config = dict(
            SWARP_CONFIG = str(self.path_global / 'swarp' / f'{self.telkey}.swarpconfig'),
            SWARP_CONFIGDIR = str(self.path_global / 'swarp'),
            SWARP_LOGDIR = str(self.path_home / 'code' / 'swarp' / 'log'),
            SWARP_HISTORYDIR = str(self.path_home / 'code' / 'swarp' / 'history')
        )
        psfex_config = dict(
            PSFEX_CONFIG = str(self.path_global / 'psfex' / 'default.psfexconfig'),
            PSFEX_SEXCONFIG = str(self.path_global / 'sextractor' / f'{self.telkey}.psfex.sexconfig'),
            PSFEX_CONFIGDIR = str(self.path_global / 'psfex'),
            PSFEX_LOGDIR = str(self.path_home / 'code' / 'psfex' / 'log'),
            PSFEX_HISTORYDIR = str(self.path_home / 'code' / 'psfex' / 'history')
        )
        
        for cfg, name in [
            (sex_config, 'sex.config'),
            (astrometry_config, 'astrometry.config'),
            (scamp_config, 'scamp.config'),
            (swarp_config, 'swarp.config'),
            (psfex_config, 'psfex.config')
        ]:
            self._make_configfile(cfg, name, savepath_tel)

        # GLOBAL CONFIGURATION
        calibdata_config = dict(
            CALIBDATA_DIR = str(self.path_home / 'data' / 'calibdata'),
            CALIBDATA_MASTERDIR = str(self.path_home / 'data' / 'mcalibdata'),
            )
        refdata_config = dict(
            REFDATA_DIR = str(self.path_home / 'data' / 'refdata'),
            )
        obsdata_config = dict(OBSDATA_DIR = str(self.path_home / 'data' / 'obsdata'))
        scidata_config = dict(SCIDATA_DIR = str(self.path_home / 'data' / 'scidata'))
        catalog_config = dict(CATALOG_DIR = str(self.path_global.parent / 'skycatalog' / 'catalog_archive'))
        observatory_config = dict(
            OBSERVATORY_LOCATIONINFO = str(self.path_global / 'obs_location.txt'),
            OBSERVATORY_TELESCOPEINFO = str(self.path_global / 'CCD.txt')
        )
        sdtdata_config = dict(
            SDTDATA_OBSSOURCEDIR = '/data/data1/obsdata/',
            SDTDATA_OBSDESTDIR = str(self.path_home / 'data' / 'obsdata' / '7DT'),
            SDTDATA_SCISOURCEDIR = '/data/data1/processed_1x1_gain2750/',
            SDTDATA_SCIDESTDIR = str(self.path_home / 'data' / 'scidata' / '7DT' / '7DT_C361K_HIGH_1x1')
        )

        for cfg, name in [
            (calibdata_config, 'calibdata.config'),
            (refdata_config, 'refdata.config'),
            (scidata_config, 'scidata.config'),
            (catalog_config, 'catalog.config'),
            (observatory_config, 'observatory.config'),
            (sdtdata_config, 'sdtdata.config'),
            (obsdata_config, 'obsdata.config')
        ]:
            self._make_configfile(cfg, name, savepath_global)

        # Remove per-telescope specific keys before saving global versions
        sex_config.pop('SEX_CONFIG', None)
        scamp_config.pop('SCAMP_SEXCONFIG', None)
        swarp_config.pop('SWARP_CONFIG', None)
        psfex_config.pop('PSFEX_SEXCONFIG', None)

        for cfg, name in [
            (sex_config, 'sex.config'),
            (scamp_config, 'scamp.config'),
            (swarp_config, 'swarp.config'),
            (psfex_config, 'psfex.config'),
        ]:
            self._make_configfile(cfg, name, savepath_global)
#%%
if __name__ == '__main__':
    telescope_keys = [
        '7DT_C361K_HIGH_1x1', '7DT_C361K_HIGH_2x2', '7DT_C361K_LOW_1x1', '7DT_C361K_LOW_2x2',
        'CBNUO_STX16803_1x1', 'LSGT_SNUCAMII_1x1', 'LSGT_ASI1600MM_1x1',
        'RASA36_KL4040_HIGH_1x1', 'RASA36_KL4040_MERGE_1x1', 'SAO_C361K_1x1',
        'SOAO_FLI4K_1x1', 'KCT_STX16803_1x1', 'SkyMapper_SG_32_Det_1x1']
    
    for key in telescope_keys:
        print(key)
        config = TIPConfig(telkey=key)
        config._initialize_config()
        print(config.config)

# %%



#%%
from pathlib import Path
import logging
import logging.handlers
import os
from typing import Union

import numpy as np
from astropy.io import fits
from astropy.time import Time
from astropy.io.fits import Header

from ezphot.configuration import TIPConfig
from ezphot.helper import Helper
#%%
class LazyFileHandler(logging.FileHandler):
    def __init__(self, filename, mode='a', encoding=None, delay=True):
        # `delay=True` avoids creating the file immediately
        super().__init__(filename, mode, encoding, delay=delay)
    
class Logger:
    def __init__(self, logger_name):
        self.path = logger_name
        self._log = self.createlogger(logger_name)

    def log(self):
        return self._log

    def createlogger(self, logger_name, logger_level='INFO'):
        logger = logging.getLogger(logger_name)
        if len(logger.handlers) > 0:
            return logger  # Logger already exists

        logger.setLevel(logger_level)
        formatter = logging.Formatter(
            datefmt='%Y-%m-%d %H:%M:%S',
            fmt='[%(levelname)s] %(asctime)-15s | %(message)s'
        )

        # Stream Handler
        streamHandler = logging.StreamHandler()
        streamHandler.setLevel(logger_level)
        streamHandler.setFormatter(formatter)
        logger.addHandler(streamHandler)

        # Lazy File Handler
        fileHandler = LazyFileHandler(filename=logger_name, delay=True)
        fileHandler.setLevel(logger_level)
        fileHandler.setFormatter(formatter)
        logger.addHandler(fileHandler)

        return logger
    
#%%
class DummyImage(TIPConfig):
    """Base class for loading, managing, and inspecting genFITS images.

    Provides lazy-loading of generated FITS data and header, WCS access, metadata extraction,
    and visualization support. Used as the parent class for all generated images
    such as Mask, Background, and ErrorMap.

    Parameters
    ----------
    path : Union[str, Path]
        Path to the FITS image file.
    """
    def __init__(self, path: Union[Path, str]):
        super().__init__()
        self.helper = Helper()
        self.path =  Path(path)
        
        self._data = None
        self._header = Header()
    
    def rename(self, new_name: str):
        """Rename the image file and update the internal path.
        
        Parameters
        ----------
        new_name : str
            New name of the image file.
        
        Returns
        -------
        None
        """
        old_path = self.path
        new_path = self.path.parent / new_name

        # If the target exists, remove it (overwrite)
        if new_path.exists():
            new_path.unlink()  # remove the existing file

        old_path.rename(new_path)
        self.path = new_path
        print(f"Renamed {old_path} to {new_path}")
    
    def clear(self):
        """Clear the image data and/or header from memory.
        
        Parameters
        ----------
        clear_data : bool, optional
            If True, clear the image data. Default is True.
        clear_header : bool, optional
            If True, clear the image header. Default is False.
        """
        self._data = None
        self._header = Header()
        print(f"Cleared data and header for {self.path}")

    def update_header(self, **kwargs):
        """Update FITS header values using known key variants.
        
        Parameters
        ----------
        kwargs : dict
            Keyword arguments to update the FITS header.
        
        Returns
        -------
        None
        """
        if self._header is None:
            print("WARNING: Header is not loaded. Cannot update.")
            return
        else:
            for key, value in kwargs.items():
                key_upper = key.upper()
                if key_upper in self._key_variants.keys():
                    key_variants = self._key_variants[key_upper]
                    for key_variant in key_variants:
                        if key_variant in self._header:
                            self._header[key_variant] = value
                else:
                    print(f'WARNING: Key {key} not found in key_variants.')

    def load(self):
        """Load both image data and header from disk.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        if not self.is_exists:
            raise FileNotFoundError(f'File not found: {self.path}')
        self.data()
        self.header()

    def show(self, 
             cmap='gray',
             scale='zscale', 
             downsample=4, 
             figsize=(8, 6), 
             title=None):
        """
        Visualize the FITS image using slicing-based downsampling and scaling.

        Parameters
        ----------
        cmap : str, optional
            Matplotlib colormap. Default is 'gray'.
        scale : str, optional
            Scaling method. Default is 'zscale'.
        downsample : int, optional
            Step size for downsampling via slicing. Default is 4.
        figsize : tuple, optional
            Matplotlib figure size. Default is (8, 6).
        title : str, optional
            Plot title. Default is None.
        save_path : str, optional
            Path to save the figure. Default is None.
        close_fig : bool, optional
            If True, close the figure after saving. Default is False.
        """
        import matplotlib.pyplot as plt
        from astropy.visualization import ZScaleInterval, MinMaxInterval
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        data = self.data
        if data is None:
            print("WARNING: Image data is not loaded. Please load the image first.")
            return
        
        # Handle NaN and inf
        data = np.nan_to_num(data, nan=0.0, posinf=0.0, neginf=0.0)

        # Store original shape before slicing
        ny, nx = data.shape

        # Downsampling using slicing
        if downsample > 1:
            data = data[::downsample, ::downsample]

        # Scaling
        if scale == 'zscale':
            interval = ZScaleInterval()
        elif scale == 'minmax':
            interval = MinMaxInterval()
        else:
            print(f"Invalid scale option: {scale}. Use 'zscale' or 'minmax'.")
            return

        vmin, vmax = interval.get_limits(data)

        # Plot image
        fig, ax = plt.subplots(figsize=figsize)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)  # colorbar axis

        img = ax.imshow(data, cmap=cmap, origin='lower', vmin=vmin, vmax=vmax)

        # Set original pixel ticks
        yticks = np.linspace(0, data.shape[0]-1, num=6, dtype=int)
        xticks = np.linspace(0, data.shape[1]-1, num=6, dtype=int)
        ax.set_yticks(yticks)
        ax.set_yticklabels((yticks * downsample).astype(int))
        ax.set_xticks(xticks)
        ax.set_xticklabels((xticks * downsample).astype(int))

        ax.set_title(title or self.path.name)
        fig.colorbar(img, cax=cax, label='Pixel value')  # colorbar on the matched axis

        plt.tight_layout()
        plt.show()
        
    def run_ds9(self):
        """Open the image with SAOImage DS9.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self.helper.run_ds9(self.path)    

    @property
    def naxis1(self):
        """Number of pixels along the first axis."""
        for key in self._key_variants['NAXIS1']:
            if key in self._header:
                return self._header[key]
        return None

    @property
    def naxis2(self):
        """Number of pixels along the second axis."""
        for key in self._key_variants['NAXIS2']:
            if key in self._header:
                return self._header[key]
        return None
    
    @property
    def exptime(self):
        """Exposure time of the image."""
        for key in self._key_variants['EXPTIME']:
            if key in self._header:
                return self._header[key]
        return None

    @property
    def egain(self):
        """Electron gain of the image."""
        for key in self._key_variants['EGAIN']:
            if key in self._header:
                return self._header[key]
        return None
            
    @property
    def wcs(self):
        """WCS information of the image."""
        try:
            from astropy.wcs import WCS
            wcs = WCS(self._header)
            return wcs
        except:
            return None
        
    @property
    def center(self):
        """Center coordinates of the image in RA and DEC."""
        center_info = dict()
        center_info['x'] = self.naxis1 / 2
        center_info['y'] = self.naxis2 / 2
        center_info['ra'] = None
        center_info['dec'] = None
        if self.wcs is not None:
            center_world = self.wcs.pixel_to_world(center_pixel[0], center_pixel[1])
            center_info['ra'] = center_world.ra.deg
            center_info['dec'] = center_world.dec.deg
        return center_info
    

    @property
    def data(self):
        """Image data array."""
        if not self.is_data_loaded and self.is_exists:
            try:
                self._data = fits.getdata(self.path)
            except Exception as e:
                print(f"Failed to load data from {self.path}: {e}")
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def header(self):
        """FITS header object."""
        if not self.is_header_loaded and self.is_exists:
            try:
                self._header = fits.getheader(self.path)
            except Exception as e:
                print(f"Failed to load header from {self.path}: {e}")
        return self._header

    @header.setter
    def header(self, value):
        self._header = value
        
    @property
    def is_data_loaded(self):
        """Check if the image data is loaded."""
        return self._data is not None

    @property
    def is_header_loaded(self):
        """Check if the image header is loaded."""
        return isinstance(self._header, Header) and len(self._header) > 0

    @property
    def is_exists(self):
        """Check if the image file exists."""
        return self.path.exists()
        
    @property
    def _key_variants(self):
        # Define key variants, if a word is duplicated in the same variant, posit the word with the highest priority first
        key_variants_upper = {
            # Default key in rawimages 
            'NAXIS1': ['NAXIS1'],
            'NAXIS2': ['NAXIS2'],
            'EGAIN': ['EGAIN'],
            'EXPTIME': ['EXPTIME', 'EXPOSURE'],
            'BINNING': ['BINNING', 'XBINNING'],
            'RDNOISE': ['RDNOISE'],
            'IMGTYPE': ['IMGTYPE'],
            'TGTPATH': ['TGTPATH'],
            'BIASPATH': ['BIASPATH'],
            'DARKPATH': ['DARKPATH'],
            'FLATPATH': ['FLATPATH'],
            # Mask
            'MASKPATH': ['MASKPATH'],
            'MASKTYPE': ['MASKTYPE'],
            'MASKATMP': ['MASKATMP'],
            # Background
            'BKGPATH': ['BKGPATH'],
            'BKGTYPE': ['BKGTYPE'],
            'BKGIS2D': ['BKGIS2D'],
            'BKGVALU': ['BKGVALU'],
            'BKGSIG': ['BKGSIG'],
            'BKGITER': ['BKGITER'],
            'BKGBOX': ['BKGBOX'],
            'BKGFILT': ['BKGFILT'],
            # Errormap
            'EMAPPATH': ['EMAPPATH'],
            'EMAPTYPE': ['EMAPTYPE'],
        }

        # Sort each list in the dictionary by string length (descending order)
        sorted_key_variants_upper = {
            key: sorted(variants, key=len, reverse=True)
            for key, variants in key_variants_upper.items()
        }
        return sorted_key_variants_upper
